import csv
import os
import sqlite3

import luigi
import orjson

from .. import _shared
from .._shared.extract import is_informative_value


class MetadataFetcher:
    _cached_values = {}

    @classmethod
    def get_versions(cls):
        # Fixed list instead of dynamically fetching it because:
        # 1) scraping Harvard Dataverse is prevented with a JavaScript challenge
        # 2) files did not change between versions 1.0, 2.0 and 2.1
        return ["2.1"]

    @classmethod
    def get_metadata(cls, version):
        # Check
        versions = cls.get_versions()
        if version not in versions:
            raise ValueError(
                f'Version "{version}" is not valid.\nAvailable options: {versions}'
            )

        # Memorize
        key = f"metadata_{version}"
        if key not in cls._cached_values:
            cls._cached_values[key] = _shared.extract.get_metadata_from_primekg()

        # Lookup
        metadata = cls._cached_values[key]
        return metadata


class FetchKnowledgeGraphFile(_shared.tasks.ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "downloads")
        self.metadata = MetadataFetcher.get_metadata(self.version)

    def requires(self):
        filename = "kg.csv.zip"
        url = self.metadata[filename]["url"]
        md5 = self.metadata[filename]["md5"]
        return _shared.tasks.DownloadFile(self.subdirpath, filename, url, md5=md5)

    def output(self):
        return self.input()


class FetchAnnotationFiles(_shared.tasks.ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "downloads")
        self.metadata = MetadataFetcher.get_metadata(self.version)
        self.filenames = [
            "disease_features.csv.zip",
            "drug_features.csv.zip",
        ]

    def requires(self):
        tasks = []
        for filename in self.filenames:
            url = self.metadata[filename]["url"]
            md5 = self.metadata[filename]["md5"]
            task = _shared.tasks.DownloadFile(self.subdirpath, filename, url, md5=md5)
            tasks.append(task)
        return tasks

    def output(self):
        return self.input()


class ExtractKnowledgeGraph(_shared.tasks.ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "downloads")

    def requires(self):
        return FetchKnowledgeGraphFile(self.dirpath, self.version)

    def output(self):
        return luigi.LocalTarget(self.input().path[:-4])

    def run(self):
        _shared.extract.extract_zip(self.input().path)


class ExtractAnnotations(_shared.tasks.ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "downloads")

    def requires(self):
        return FetchAnnotationFiles(self.dirpath, self.version)

    def output(self):
        return [luigi.LocalTarget(target.path[:-4]) for target in self.input()]

    def run(self):
        for target in self.input():
            _shared.extract.extract_zip(target.path)


class CreateSqliteFile(_shared.tasks.ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()
    batch_size = luigi.IntParameter(default=10_000, significant=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "results")
        self.filepath = os.path.join(self.subdirpath, "kg.sqlite")

    def requires(self):
        tasks = {
            "subdirpath": _shared.tasks.CreateDirectory(self.subdirpath),
            "kg_file": ExtractKnowledgeGraph(self.dirpath, self.version),
            "annotation_files": ExtractAnnotations(self.dirpath, self.version),
        }
        return tasks

    def output(self):
        return luigi.LocalTarget(self.filepath)

    def run(self):
        with _shared.tasks.temp_output_path(self.filepath) as sqlite_filepath:
            with sqlite3.connect(sqlite_filepath) as conn:
                cursor = conn.cursor()
                self._prepare_annotations()
                _shared.transform.create_sql_schema(cursor)
                self._insert_nodes(conn, cursor)
                self._insert_edges(conn, cursor)
                conn.commit()
        self._remove_input_csv_files()

    def _prepare_annotations(self):
        node_id_to_annotation_map = {}
        for target in self.input()["annotation_files"]:
            filepath = target.path

            with open(filepath, newline="") as f:
                reader = csv.reader(f, delimiter=",")
                header = next(reader)
                data_columns = header[1:]
                for row in reader:
                    node_id = row[0]
                    annotation = {}
                    for i, col in enumerate(data_columns, 1):
                        value = row[i]
                        if is_informative_value(value):
                            annotation[col] = value
                    node_id_to_annotation_map[node_id] = annotation
        self.node_id_to_annotation_map = node_id_to_annotation_map

    def _insert_nodes(self, conn, cursor):
        # Special aspects in handling the raw data
        # 1) Node property "id" is renamed to "identifier" to prevent an issue with GraphML.
        kg_filepath = self.input()["kg_file"].path
        sql_cmd = """
            INSERT INTO nodes (id, type, properties)
            VALUES (?, ?, ?)
        """
        seen_node_ids = set()
        with open(kg_filepath, "r", newline="") as csvfile:
            reader = csv.reader(csvfile, delimiter=",")
            columns = next(reader)
            x_index_idx = columns.index("x_index")
            x_id_idx = columns.index("x_id")
            x_type_idx = columns.index("x_type")
            x_name_idx = columns.index("x_name")
            x_source_idx = columns.index("x_source")
            y_index_idx = columns.index("y_index")
            y_id_idx = columns.index("y_id")
            y_type_idx = columns.index("y_type")
            y_name_idx = columns.index("y_name")
            y_source_idx = columns.index("y_source")
            batch = []
            for row in reader:
                x_index = row[x_index_idx]
                x_id = row[x_id_idx]
                x_type = row[x_type_idx]
                x_name = row[x_name_idx]
                x_source = row[x_source_idx]
                y_index = row[y_index_idx]
                y_id = row[y_id_idx]
                y_type = row[y_type_idx]
                y_name = row[y_name_idx]
                y_source = row[y_source_idx]
                # x
                node_id = x_index
                if node_id not in seen_node_ids:
                    seen_node_ids.add(node_id)
                    node_type = x_type
                    node_properties = {
                        "identifier": x_id,
                        "name": x_name,
                        "source": x_source,
                    }
                    if node_id in self.node_id_to_annotation_map:
                        node_properties.update(self.node_id_to_annotation_map[node_id])
                    node_properties_str = orjson.dumps(node_properties).decode("utf-8")
                    node = (node_id, node_type, node_properties_str)
                    batch.append(node)
                # y
                node_id = y_index
                if node_id not in seen_node_ids:
                    seen_node_ids.add(node_id)
                    node_type = y_type
                    node_properties = {
                        "identifier": y_id,
                        "name": y_name,
                        "source": y_source,
                    }
                    node_properties = {
                        key: val
                        for key, val in node_properties.items()
                        if is_informative_value(val)
                    }
                    if node_id in self.node_id_to_annotation_map:
                        node_properties.update(self.node_id_to_annotation_map[node_id])
                    node_properties_str = orjson.dumps(node_properties).decode("utf-8")
                    node = (node_id, node_type, node_properties_str)
                    batch.append(node)
                if len(batch) >= self.batch_size:
                    cursor.executemany(sql_cmd, batch)
                    batch = []
            if batch:
                cursor.executemany(sql_cmd, batch)
        conn.commit()

    def _insert_edges(self, conn, cursor):
        kg_filepath = self.input()["kg_file"].path
        sql_cmd = """
            INSERT INTO edges (source_id, target_id, type, properties)
            VALUES (?, ?, ?, ?)
        """
        with open(kg_filepath, "r", newline="") as csvfile:
            reader = csv.reader(csvfile, delimiter=",")
            columns = next(reader)
            relation_idx = columns.index("relation")
            display_relation_idx = columns.index("display_relation")
            x_index_idx = columns.index("x_index")
            y_index_idx = columns.index("y_index")
            batch = []
            for row in reader:
                source_id = row[x_index_idx]
                target_id = row[y_index_idx]
                edge_type = row[relation_idx]
                edge_properties = {"display_relation": row[display_relation_idx]}
                edge_properties = {
                    key: val
                    for key, val in edge_properties.items()
                    if is_informative_value(val)
                }
                edge_properties_str = orjson.dumps(edge_properties).decode("utf-8")
                edge = (source_id, target_id, edge_type, edge_properties_str)
                batch.append(edge)
                if len(batch) >= self.batch_size:
                    cursor.executemany(sql_cmd, batch)
                    batch = []
            if batch:
                cursor.executemany(sql_cmd, batch)
        conn.commit()

    def _remove_input_csv_files(self):
        kg_filepath = self.input()["kg_file"].path
        annotation_filepaths = [
            target.path for target in self.input()["annotation_files"]
        ]
        all_filepaths = [kg_filepath] + annotation_filepaths
        for filepath in all_filepaths:
            try:
                os.remove(filepath)
            except Exception:
                pass


class PrimeKg(_shared.base.Project):
    """Precision Medicine Knowledge Graph (PrimeKG).

    References
    ----------
    - Publication: https://doi.org/10.1038/s41597-023-01960-3
    - Website: https://zitniklab.hms.harvard.edu/projects/PrimeKG
    - Code: https://github.com/mims-harvard/PrimeKG
    - Data: https://doi.org/10.7910/DVN/IXA7BM

    """

    _label = "primekg"
    _MetadataFetcher = MetadataFetcher
    _CreateSqliteFile = CreateSqliteFile
    _node_type_to_color = {
        "drug": "green",
        "gene/protein": "blue",
        "disease": "red",
        "pathway": "red",
        "biological_process": "red",
    }



import csv
import os
import sqlite3

import luigi
import orjson

from .. import _shared
from .._shared.extract import is_informative_value


class MetadataFetcher:
    dataset_id = "mrcf7f4tc2"
    _cached_values = {}

    @classmethod
    def get_versions(cls):
        # Memorize
        key = "versions"
        if key not in cls._cached_values:
            cls._cached_values[key] = _shared.extract.get_versions_from_mendeley(
                cls.dataset_id
            )

        # Lookup
        versions = cls._cached_values[key]
        return versions

    @classmethod
    def get_metadata(cls, version):
        # Check
        versions = cls.get_versions()
        if version not in versions:
            raise ValueError(
                f'Version "{version}" is not valid.\nAvailable options: {versions}'
            )

        # Memorize
        key = f"metadata_{version}"
        if key not in cls._cached_values:
            cls._cached_values[key] = _shared.extract.get_metadata_from_mendeley(
                cls.dataset_id, version
            )

        # Lookup
        metadata = cls._cached_values[key]
        return metadata


class FetchNeo4jDumpFile(_shared.tasks.ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "downloads")
        self.metadata = MetadataFetcher.get_metadata(self.version)

    def _detect_filename(self):
        # The filename is slightly different for different versions, hence it has to be detected
        for candidate in self.metadata:
            if candidate.endswith(".dump"):
                return candidate
        raise ValueError(
            "Could not identify the Neo4j dump file that contains the knowledge graph."
        )

    def requires(self):
        filename = self._detect_filename()
        url = self.metadata[filename]["url"]
        sha256 = self.metadata[filename]["sha256"]
        return _shared.tasks.DownloadFile(self.subdirpath, filename, url, sha256=sha256)

    def output(self):
        return self.input()


class FetchNeo4jApocFile(_shared.tasks.ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "downloads")

    def requires(self):
        filename = "apoc-4.4.0.24-all.jar"
        url = (
            "https://github.com/neo4j-contrib/neo4j-apoc-procedures/releases/"
            "download/4.4.0.24/apoc-4.4.0.24-all.jar"
        )
        md5 = "7c6a702322b0aaf663c25f378cd3494d"
        return _shared.tasks.DownloadFile(self.subdirpath, filename, url, md5=md5)

    def output(self):
        return self.input()


class ExportNeo4jToCsvFiles(_shared.tasks.ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "results")

    def requires(self):
        return {
            "neo4j_database_dump": FetchNeo4jDumpFile(self.dirpath, self.version),
            "neo4j_apoc_plugin": FetchNeo4jApocFile(self.dirpath, self.version),
        }

    def output(self):
        return {
            "nodes_file": luigi.LocalTarget(
                os.path.join(self.subdirpath, "neo4j_nodes.csv")
            ),
            "edges_file": luigi.LocalTarget(
                os.path.join(self.subdirpath, "neo4j_edges.csv")
            ),
        }

    def run(self):
        dw = DockerWrapper()

        # Check docker installation
        if not dw.test_installation():
            msg = (
                "Testing the Docker installation on your system indicated that "
                "it does not work as expected on a minimal example. A potential "
                "issue can be a missing internet connection for fetching remote "
                "Docker images."
            )
            raise ValueError(msg)

        # Convert .dump to .csv
        filepath_db = self.input()["neo4j_database_dump"].path
        filepath_apoc = self.input()["neo4j_apoc_plugin"].path
        dirpath_source = os.path.dirname(filepath_db)
        dirpath_target = self.subdirpath
        filename_db = os.path.basename(filepath_db)
        filename_apoc = os.path.basename(filepath_apoc)
        dw.convert_neo4j_dump_to_csv_files(
            dirpath_source, dirpath_target, filename_db, filename_apoc
        )


class CreateSqliteFile(_shared.tasks.ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()
    batch_size = luigi.IntParameter(default=10_000, significant=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "results")
        self.filepath = os.path.join(self.subdirpath, "kg.sqlite")

    def requires(self):
        tasks = {
            "subdirpath": _shared.tasks.CreateDirectory(self.subdirpath),
            "csv_files": ExportNeo4jToCsvFiles(self.dirpath, self.version),
        }
        return tasks

    def output(self):
        return luigi.LocalTarget(self.filepath)

    def run(self):
        with _shared.tasks.temp_output_path(self.filepath) as sqlite_filepath:
            with sqlite3.connect(sqlite_filepath) as conn:
                cursor = conn.cursor()
                _shared.transform.create_sql_schema(cursor)
                self._insert_nodes(conn, cursor)
                self._insert_edges(conn, cursor)
                conn.commit()
        self._remove_input_csv_files()

    def _insert_nodes(self, conn, cursor):
        nodes_filepath = self.input()["csv_files"]["nodes_file"].path
        sql_cmd = """
            INSERT INTO nodes (id, type, properties)
            VALUES (?, ?, ?)
        """
        with open(nodes_filepath, "r", newline="") as f:
            reader = csv.reader(f, delimiter=",")
            columns = next(reader)
            batch = []
            for row in reader:
                node_id = str(row[0])
                node_type = str(row[1])
                node_properties = {
                    str(key): val
                    for key, val in orjson.loads(row[2]).items()
                    if is_informative_value(val)
                }
                node_properties_str = orjson.dumps(node_properties).decode("utf-8")
                node = (node_id, node_type, node_properties_str)
                batch.append(node)
                if len(batch) >= self.batch_size:
                    cursor.executemany(sql_cmd, batch)
                    batch = []
            if batch:
                cursor.executemany(sql_cmd, batch)
        conn.commit()

    def _insert_edges(self, conn, cursor):
        edges_filepath = self.input()["csv_files"]["edges_file"].path
        sql_cmd = """
            INSERT INTO edges (source_id, target_id, type, properties)
            VALUES (?, ?, ?, ?)
        """
        with open(edges_filepath, "r", newline="") as f:
            reader = csv.reader(f, delimiter=",")
            columns = next(reader)
            batch = []
            for row in reader:
                source_id = str(row[0])
                target_id = str(row[1])
                edge_type = str(row[2])
                edge_properties = {
                    str(key): val
                    for key, val in orjson.loads(row[3]).items()
                    if is_informative_value(val)
                }
                edge_properties_str = orjson.dumps(edge_properties).decode("utf-8")
                edge = (source_id, target_id, edge_type, edge_properties_str)
                batch.append(edge)
                if len(batch) >= self.batch_size:
                    cursor.executemany(sql_cmd, batch)
                    batch = []
            if batch:
                cursor.executemany(sql_cmd, batch)
        conn.commit()

    def _remove_input_csv_files(self):
        nodes_filepath = self.input()["csv_files"]["nodes_file"].path
        edges_filepath = self.input()["csv_files"]["edges_file"].path
        for filepath in (nodes_filepath, edges_filepath):
            try:
                os.remove(filepath)
            except Exception:
                pass


class DockerWrapper:
    def __init__(self):
        try:
            import docker

            self._docker_module = docker
        except ImportError:
            msg = (
                "This knowledge graph project requires 1) a Docker installation on "
                'your operating system and 2) the Python package "docker" for the '
                "language bindings, which can be installed with `pip install docker`."
            )
            raise ImportError(msg)

    def test_installation(self):
        try:
            # Definition
            container_name = "hello-world"
            expected_message = b"Hello from Docker!"
            # Run
            client = self._docker_module.from_env()
            client.images.pull(container_name)
            container = client.containers.run(container_name, detach=True)
            works = expected_message in container.logs()
            # Cleanup
            container.remove(force=True)
            client.images.remove(container_name, force=True)
            client.containers.prune()
            client.images.prune()
            client.volumes.prune()
        except Exception:
            works = False
        return works

    def run_command(self, container, command):
        print(command)
        exit_code, output = container.exec_run(command)
        result = output.decode().strip()
        print(result)
        print()
        return result

    def stop_and_remove_container(self, name):
        try:
            client = self._docker_module.from_env()
            container = client.containers.get(name)
            if container.status == "running":
                container.stop()
            container.remove(force=True)
            client.containers.prune()
            client.images.prune()
            client.volumes.prune()
        except Exception:
            pass

    def convert_neo4j_dump_to_csv_files(
        self, dirpath_source, dirpath_target, filename_db, filename_apoc
    ):
        # APOC library provides CSV export: https://neo4j.com/labs/apoc/4.4/overview/apoc.export

        # Temporary directory for output to prevent side effects such as changed permissions
        dirpath_target_temp = os.path.join(dirpath_target, "tempdir_for_docker")
        os.makedirs(dirpath_target_temp, exist_ok=True)

        # Fetch and run the Neo4j v4.4 Docker container
        unique_container_name = "kgw_neo4j_198bc9da7d"
        self.stop_and_remove_container(unique_container_name)
        client = self._docker_module.from_env()
        container = client.containers.run(
            image="neo4j:4.4",
            name=unique_container_name,
            detach=True,
            volumes={
                os.path.abspath(dirpath_source): {"bind": "/source", "mode": "rw"},
                os.path.abspath(dirpath_target_temp): {
                    "bind": "/target",
                    "mode": "rw",
                },
            },
            working_dir="/target",
            entrypoint="tail -f /dev/null",
        )

        # Convert the .dump file to two .csv files
        commands = [
            # Configure Neo4j for the import
            '''sh -c "echo 'dbms.allow_upgrade=true' >> /var/lib/neo4j/conf/neo4j.conf"''',
            '''sh -c "echo 'dbms.security.procedures.allowlist=apoc.*' >> /var/lib/neo4j/conf/neo4j.conf"''',
            '''sh -c "echo 'dbms.security.procedures.unrestricted=apoc.*' >> /var/lib/neo4j/conf/neo4j.conf"''',
            '''sh -c "echo 'apoc.export.file.enabled=true' >> /var/lib/neo4j/conf/neo4j.conf"''',
            '''sh -c "neo4j-admin set-initial-password correcthorsebatterystaple"''',
            # Make APOC extension available to Neo4j
            f'''sh -c "cp /source/{filename_apoc} /var/lib/neo4j/plugins"''',
            # Import the .dump file
            f'''sh -c "neo4j-admin load --from=/source/{filename_db} --database=neo4j --force"''',
            # Start the database
            '''sh -c "rm -rf /var/lib/neo4j/logs && neo4j start"''',
            # Wait so it is ready
            '''sh -c "sleep 360"''',
            # Export a .csv file for nodes
            '''cypher-shell -u neo4j -p correcthorsebatterystaple -d neo4j "CALL apoc.export.csv.query('MATCH (n) RETURN id(n) as id, labels(n)[0] as type, properties(n) as properties', 'nodes.csv', {})"''',
            # Export a .csv file for edges
            '''cypher-shell -u neo4j -p correcthorsebatterystaple -d neo4j "CALL apoc.export.csv.query('MATCH ()-[r]->() RETURN id(startNode(r)) as source_id, id(endNode(r)) as target_id, type(r) as type, properties(r) as properties', 'edges.csv', {})"''',
            # Move the .csv files to the mounted directory to make them available outside the container
            '''sh -c "mv /var/lib/neo4j/import/nodes.csv /target"''',
            '''sh -c "mv /var/lib/neo4j/import/edges.csv /target"''',
        ]
        for cmd in commands:
            self.run_command(container, cmd)

        # Remove the container
        self.stop_and_remove_container(unique_container_name)

        # Move CSV files from temporary to target directory
        os.rename(
            os.path.join(dirpath_target_temp, "nodes.csv"),
            os.path.join(dirpath_target, "neo4j_nodes.csv"),
        )
        os.rename(
            os.path.join(dirpath_target_temp, "edges.csv"),
            os.path.join(dirpath_target, "neo4j_edges.csv"),
        )
        os.rmdir(dirpath_target_temp)
        return container


class Ckg(_shared.base.Project):
    """Clinical Knowledge Graph (CKG).

    References
    ----------
    - Publication: https://doi.org/10.1038/s41587-021-01145-6
    - Website: https://ckg.readthedocs.io
    - Code: https://github.com/MannLabs/CKG
    - Data: https://doi.org/10.17632/mrcf7f4tc2

    """

    _label = "ckg"
    _MetadataFetcher = MetadataFetcher
    _CreateSqliteFile = CreateSqliteFile
    _node_type_to_color = {
        "Metabolite": "green",
        "Known_variant": "blue",
        "Gene": "blue",
        "Transcript": "blue",
        "Protein": "blue",
        "Modified_protein": "blue",
        "Peptide": "blue",
        "Disease": "red",
        "Pathway": "red",
        "Biological_process": "red",
    }



import csv
import os
import sqlite3

import luigi
import orjson

from .. import _shared
from .._shared.extract import is_informative_value


class MetadataFetcher:
    _cached_values = {}

    @classmethod
    def get_versions(cls):
        # Memorize
        key = "versions"
        if key not in cls._cached_values:
            cls._cached_values[key] = _shared.extract.get_versions_from_monarch()

        # Lookup
        versions = cls._cached_values[key]
        return versions

    @classmethod
    def get_metadata(cls, version):
        # Check
        versions = cls.get_versions()
        if version not in versions:
            raise ValueError(
                f'Version "{version}" is not valid.\nAvailable options: {versions}'
            )

        # Memorize
        key = f"metadata_{version}"
        if key not in cls._cached_values:
            cls._cached_values[key] = _shared.extract.get_metadata_from_monarch(version)

        # Lookup
        metadata = cls._cached_values[key]
        return metadata


class FetchKnowledgeGraphFile(_shared.tasks.ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "downloads")
        self.metadata = MetadataFetcher.get_metadata(self.version)

    def requires(self):
        filename = "monarch-kg.tar.gz"
        url = self.metadata[filename]["url"]
        return _shared.tasks.DownloadFile(self.subdirpath, filename, url)

    def output(self):
        return self.input()


class DecompressKnowledgeGraphFile(_shared.tasks.ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "downloads")

    def requires(self):
        return FetchKnowledgeGraphFile(self.dirpath, self.version)

    def output(self):
        return {
            "nodes_file": luigi.LocalTarget(
                os.path.join(self.subdirpath, "monarch-kg_nodes.tsv")
            ),
            "edges_file": luigi.LocalTarget(
                os.path.join(self.subdirpath, "monarch-kg_edges.tsv")
            ),
        }

    def run(self):
        _shared.extract.extract_tar_gz(self.input().path)


class CreateSqliteFile(_shared.tasks.ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()
    batch_size = luigi.IntParameter(default=10_000, significant=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "results")
        self.filepath = os.path.join(self.subdirpath, "kg.sqlite")

    def requires(self):
        tasks = {
            "subdirpath": _shared.tasks.CreateDirectory(self.subdirpath),
            "csv_files": DecompressKnowledgeGraphFile(self.dirpath, self.version),
        }
        return tasks

    def output(self):
        return luigi.LocalTarget(self.filepath)

    def run(self):
        with _shared.tasks.temp_output_path(self.filepath) as sqlite_filepath:
            with sqlite3.connect(sqlite_filepath) as conn:
                cursor = conn.cursor()
                _shared.transform.create_sql_schema(cursor)
                self._insert_nodes(conn, cursor)
                self._insert_edges(conn, cursor)
                conn.commit()
        self._remove_decompressed_input_files()

    def _insert_nodes(self, conn, cursor):
        nodes_filepath = self.input()["csv_files"]["nodes_file"].path
        sql_cmd = """
            INSERT INTO nodes (id, type, properties)
            VALUES (?, ?, ?)
        """
        skipped_keys = ["id", "category"]
        with open(nodes_filepath, "r", newline="") as f:
            reader = csv.reader(f, delimiter="\t")
            columns = next(reader)
            id_idx = columns.index("id")
            type_idx = columns.index("category")
            properties_name_idx = [
                (col, columns.index(col)) for col in columns if col not in skipped_keys
            ]
            batch = []
            for row in reader:
                node_id = row[id_idx]
                node_type = row[type_idx]
                node_properties = {
                    name: row[idx]
                    for name, idx in properties_name_idx
                    if is_informative_value(row[idx])
                }
                node_properties_str = orjson.dumps(node_properties).decode("utf-8")
                node = (node_id, node_type, node_properties_str)
                batch.append(node)
                if len(batch) >= self.batch_size:
                    cursor.executemany(sql_cmd, batch)
                    batch = []
            if batch:
                cursor.executemany(sql_cmd, batch)
        conn.commit()

    def _insert_edges(self, conn, cursor):
        edges_filepath = self.input()["csv_files"]["edges_file"].path
        sql_cmd = """
            INSERT INTO edges (source_id, target_id, type, properties)
            VALUES (?, ?, ?, ?)
        """
        skipped_keys = ["subject", "object", "predicate"]
        with open(edges_filepath, "r", newline="") as f:
            reader = csv.reader(f, delimiter="\t")
            columns = next(reader)
            source_id_idx = columns.index("subject")
            target_id_idx = columns.index("object")
            type_idx = columns.index("predicate")
            properties_name_idx = [
                (col, columns.index(col)) for col in columns if col not in skipped_keys
            ]
            batch = []
            for row in reader:
                source_id = row[source_id_idx]
                target_id = row[target_id_idx]
                edge_type = row[type_idx]
                edge_properties = {
                    name: row[idx]
                    for name, idx in properties_name_idx
                    if is_informative_value(row[idx])
                }
                edge_properties_str = orjson.dumps(edge_properties).decode("utf-8")
                edge = (source_id, target_id, edge_type, edge_properties_str)
                batch.append(edge)
                if len(batch) >= self.batch_size:
                    cursor.executemany(sql_cmd, batch)
                    batch = []
            if batch:
                cursor.executemany(sql_cmd, batch)
        conn.commit()

    def _remove_decompressed_input_files(self):
        nodes_filepath = self.input()["csv_files"]["nodes_file"].path
        edges_filepath = self.input()["csv_files"]["edges_file"].path
        for filepath in (nodes_filepath, edges_filepath):
            try:
                os.remove(filepath)
            except Exception:
                pass


class MonarchKg(_shared.base.Project):
    """Monarch Knowledge Graph (MonarchKG).

    References
    ----------
    - Publication: https://doi.org/10.1093/nar/gkad1082
    - Website: https://monarchinitiative.org
    - Code: https://github.com/monarch-initiative/monarch-ingest
    - Data: https://data.monarchinitiative.org/monarch-kg/index.html

    """

    _label = "monarchkg"
    _MetadataFetcher = MetadataFetcher
    _CreateSqliteFile = CreateSqliteFile
    _node_type_to_color = {
        "biolink:Drug": "green",
        "biolink:ChemicalEntity": "green",
        "biolink:MolecularEntity": "green",
        "biolink:SmallMolecule": "green",
        "biolink:Gene": "blue",
        "biolink:Protein": "blue",
        "biolink:Disease": "red",
        "biolink:Pathway": "red",
        "biolink:BiologicalProcessOrActivity": "red",
    }



import csv
import json
import os
import sys

import igraph as ig
import networkx as nx


def get_project_dir(name, version):
    return f"{name}_v{version}"


def check_file(filepath):
    assert os.path.isfile(filepath), f'File "{filepath}" does not exist'
    assert os.path.getsize(filepath) > 0, f'File "{filepath}" is empty'


def load_text_file(filepath):
    with open(filepath) as f:
        data = f.read()
    return data


def load_json_file(filepath):
    with open(filepath) as f:
        data = json.load(f)
    assert data
    return data


def load_jsonl_file(filepath):
    data = []
    with open(filepath) as file:
        for line in file:
            assert len(line) > 1, "Blank line"
            item = json.loads(line)
            data.append(item)
    return data


def load_csv_file(filepath):
    csv.field_size_limit(sys.maxsize)
    with open(filepath, "r", newline="") as csvfile:
        reader = csv.reader(csvfile)
        columns = next(reader)
        data = []
        for row in reader:
            assert len(row) > 0, "Blank line"
            data.append(row)

    assert isinstance(data, list)
    assert isinstance(columns, list)
    assert len(data) > 0
    assert len(columns) > 0
    assert len(data[0]) == len(columns)
    return data, columns


def generate_sample_indices(num_total, num_samples=42):
    return [int(num_total / num_samples * i) for i in range(num_samples)]


def load_graphml_igraph(filepath):
    # Parsing GraphML
    graph = ig.Graph.Read_GraphML(filepath)
    # Counts
    num_nodes = graph.vcount()
    num_edges = graph.ecount()
    # Sample nodes
    sample_nodes_idx = generate_sample_indices(num_nodes)
    sample_nodes = [graph.vs[idx] for idx in sample_nodes_idx]
    skip_attrs = ["id", "node_type"]
    parsed_nodes = []
    for chosen_node in sample_nodes:
        node_id = chosen_node["id"]
        node_type = chosen_node["node_type"]
        node_attributes = graph.vertex_attributes()
        node_properties = {
            attr: chosen_node[attr]
            for attr in node_attributes
            if attr not in skip_attrs
        }
        node_properties = {
            k: v
            for k, v in sorted(node_properties.items())
            if v is not None and v != ""
        }
        parsed_nodes.append((node_id, node_type, node_properties))
    return num_nodes, num_edges, parsed_nodes


def load_graphml_networkx(filepath):
    # Parsing GraphML
    graph = nx.read_graphml(filepath)
    # Counts
    num_nodes = graph.number_of_nodes()
    num_edges = graph.number_of_edges()
    # Sample nodes
    sample_nodes_idx = generate_sample_indices(num_nodes)
    sample_nodes = [
        node
        for idx, node in enumerate(graph.nodes(data=True))
        if idx in sample_nodes_idx
    ]
    skip_attrs = ["id", "node_type"]
    parsed_nodes = []
    for chosen_node in sample_nodes:
        node_id, node_data = chosen_node
        node_type = node_data["node_type"]
        node_properties = {
            k: v for k, v in sorted(node_data.items()) if k not in skip_attrs
        }
        parsed_nodes.append((node_id, node_type, node_properties))
    return num_nodes, num_edges, parsed_nodes



import os

import pytest

import kgw


# Fixtures are automatically recognized by pytest, no need to import them


@pytest.fixture(scope="session")
def workdir():
    workdir = "test_kgw_workdir"

    # Setup: remove the directory if it already exists
    # Commented out to enable reuse of partial results
    # if os.path.exists(workdir):
    #    shutil.rmtree(workdir)

    # Create a fresh directory
    os.makedirs(workdir, exist_ok=True)

    # Yield control to the test
    yield workdir

    # Teardown: remove the directory after the test completes
    # Commented out to enable reuse and sharing of results
    # if os.path.exists(workdir):
    #    shutil.rmtree(workdir)


@pytest.fixture(
    params=[
        (
            "hald",
            kgw.biomedicine.Hald,
            kgw.biomedicine.Hald.get_versions()[-1],
            12_257,
            116_495,
        ),
        (
            "oregano",
            kgw.biomedicine.Oregano,
            kgw.biomedicine.Oregano.get_versions()[-1],
            88_937,
            819_884,
        ),
        (
            "primekg",
            kgw.biomedicine.PrimeKg,
            kgw.biomedicine.PrimeKg.get_versions()[-1],
            129_375,
            8_100_498,
        ),
        (
            "monarchkg",
            kgw.biomedicine.MonarchKg,
            kgw.biomedicine.MonarchKg.get_versions()[-1],
            1_028_155,
            11_076_689,
        ),
        (
            "ckg",
            kgw.biomedicine.Ckg,
            kgw.biomedicine.Ckg.get_versions()[-1],
            14_543_042,
            188_266_233,
        ),
    ]
)
def project_params_full(request):
    return request.param


@pytest.fixture(
    params=[
        (
            "hald",
            kgw.biomedicine.Hald,
            kgw.biomedicine.Hald.get_versions()[-1],
            12_257,
            116_495,
        ),
        (
            "oregano",
            kgw.biomedicine.Oregano,
            kgw.biomedicine.Oregano.get_versions()[-1],
            88_937,
            819_884,
        ),
    ]
)
def project_params(request):
    return request.param



import os

import pytest
import requests

from kgw import _shared as shared


@pytest.mark.parametrize(
    "command, expected_result",
    [
        ("echo 'hi there'", "hi there"),
        ("python --version", "."),
        ("docker --help", "docker"),
    ],
)
def test_run_shell_command(command, expected_result):
    result = shared.extract.run_shell_command(command)
    assert expected_result in result


def test_create_dir():
    dirname = "some_new_directory"
    shared.extract.create_dir(dirname)
    assert os.path.isdir(dirname)
    os.rmdir(dirname)


def test_get_file_size_and_delete_file():
    filename = "some_new_file"
    shared.extract.run_shell_command(f"echo 'hello world' > {filename}")
    assert os.path.isfile(filename)
    assert shared.extract.get_file_size(filename) > 0
    shared.extract.delete_file(filename)
    assert not os.path.isfile(filename)


def test_has_internet_connection_and_get_request_with_retries():
    assert shared.extract.has_internet_connection(), "No internet connection"

    shared.extract.ensure_internet_connection()

    url = "https://www.google.com"
    response = shared.extract.get_request_with_retries(url)
    assert "<" in response.text and ">" in response.text

    url = "https://www.somenonsense-urllajsfnalfkjajf1231231knsdfölsf.at"

    with pytest.raises(requests.exceptions.ConnectionError):
        shared.extract.get_request_with_retries(url, num_retries=1, delay_in_sec=0.1)


def test_get_remote_file_size():
    url = "https://www.google.com/robots.txt"
    assert shared.extract.get_remote_file_size(url) > 0

    url = "https://www.google.com/nonsensicalfilethatdoesnotexist12319182391.txt"
    assert shared.extract.get_remote_file_size(url) == 0


def test_download_file_and_fetch_file():
    url = "https://www.google.com/robots.txt"
    filename = "stobor_elgoog.txt"
    for func in (shared.extract.download_file, shared.extract.fetch_file):
        func(url, filename)
        assert os.path.isfile(filename)
        assert shared.extract.get_file_size(filename) > 0
        shared.extract.delete_file(filename)
        assert not os.path.isfile(filename)


def test_is_informative_value():
    for val in ["a", 123, 3.14, ("a", "b"), [1, 2], -1]:
        assert shared.extract.is_informative_value(val)

    for val in [None, "", float("nan")]:
        assert not shared.extract.is_informative_value(val)


def test_clean():
    assert shared.load.clean(1) == "1"
    assert shared.load.clean(3.14) == "3.14"
    assert shared.load.clean("a c") == '"a c"'
    assert shared.load.clean([1, 3.14, "a c"]) == r'"[1,3.14,\"a c\"]"'



"""Knowledge graph projects from the domain of biomedicine."""

from ._ckg import Ckg
from ._hald import Hald
from ._monarchkg import MonarchKg
from ._oregano import Oregano
from ._primekg import PrimeKg


__all__ = [
    "Ckg",
    "Hald",
    "MonarchKg",
    "Oregano",
    "PrimeKg",
]



"""kgw - Knowledge Graph Workflows

A Python package for downloading, converting, and analyzing
a selection of knowledge graphs.

Currently five projects from the domain of biomedicine
are covered. In future, more projects from the same or other
domains might get added. Contributions are welcome and encouraged!

"""

__version__ = "0.1.0"

__all__ = [
    "biomedicine",
    "run",
]


from . import biomedicine
from ._shared.base import run



import csv
import sqlite3
import xml.sax.saxutils as saxutils

import gravis as gv
import networkx as nx
import orjson


def sqlite_to_statistics(db_filepath, statistics_filepath):
    with sqlite3.connect(db_filepath) as conn:
        cursor = conn.cursor()

        # Node statistics
        query = """
            SELECT
                (SELECT COUNT(*) FROM nodes) AS num_nodes,
                (SELECT COUNT(DISTINCT type) FROM nodes) AS num_node_types,
                json_group_object(type, count) AS node_types
            FROM (
                SELECT
                    type,
                    COUNT(*) AS count
                FROM
                    nodes
                GROUP BY
                    type
                ORDER BY
                    count DESC,
                    type ASC
            )
        """
        cursor.execute(query)
        row = cursor.fetchone()
        node_data = {
            "num_nodes": row[0],
            "num_node_types": row[1],
            "node_types": orjson.loads(row[2]),
        }

        # Edge statistics
        query = """
            SELECT
                (SELECT COUNT(*) FROM edges) AS num_edges,
                (SELECT COUNT(DISTINCT type) FROM edges) AS num_edge_types,
                json_group_object(type, count) AS edge_types
            FROM (
                SELECT
                    type,
                    COUNT(*) AS count
                FROM
                    edges
                GROUP BY
                    type
                ORDER BY
                    count DESC,
                    type ASC
            )
        """
        cursor.execute(query)
        row = cursor.fetchone()
        edge_data = {
            "num_edges": row[0],
            "num_edge_types": row[1],
            "edge_types": orjson.loads(row[2]),
        }

        # Combination and reordering
        data = {
            "num_nodes": node_data["num_nodes"],
            "num_edges": edge_data["num_edges"],
            "num_node_types": node_data["num_node_types"],
            "num_edge_types": edge_data["num_edge_types"],
            "node_types": node_data["node_types"],
            "edge_types": edge_data["edge_types"],
        }

        # Writing to file
        with open(statistics_filepath, "wb") as f:
            f.write(orjson.dumps(data, option=orjson.OPT_INDENT_2))
    return data


def sqlite_to_schema(db_filepath, schema_filepath, node_type_to_color=None):
    if node_type_to_color is None:
        node_type_to_color = {}

    with sqlite3.connect(db_filepath) as conn:
        cursor = conn.cursor()

        # Nodes
        query = """
            SELECT
                type,
                COUNT(*) as count
            FROM
                nodes
            GROUP BY
                type
            ORDER BY
                count DESC,
                type ASC
        """
        cursor.execute(query)
        rows = cursor.fetchall()
        node_data = list(rows)

        # Edges
        query = """
            SELECT
                source_node.type AS source_type,
                edges.type AS edge_type,
                target_node.type AS target_type,
                COUNT(*) AS triple_count
            FROM
                edges
            JOIN
                nodes AS source_node
            ON
                edges.source_id = source_node.id
            JOIN
                nodes AS target_node
            ON
                edges.target_id = target_node.id
            GROUP BY
                source_type, edge_type, target_type
            ORDER BY
                triple_count DESC
        """
        cursor.execute(query)
        rows = cursor.fetchall()
        edge_data = list(rows)

        # Graph
        g = nx.DiGraph()
        for node_type, cnt_instances in node_data:
            size = cnt_instances
            color = node_type_to_color.get(node_type, None)
            hover = (
                f"Node type: {node_type}\n\n" f"Number of instances: {cnt_instances}"
            )
            g.add_node(
                node_type, size=size, color=color, label_color=color, hover=hover
            )

        for source_type, edge_type, target_type, cnt_instances in edge_data:
            size = cnt_instances
            color = node_type_to_color.get(source_type, None)
            hover = (
                f"Edge type: {edge_type}\n"
                f"Source: {source_type}\n"
                f"Target: {target_type})\n\n"
                f"Number of instances: {cnt_instances}"
            )
            g.add_edge(source_type, target_type, size=size, color=color, hover=hover)

        # Plot
        fig = gv.d3(
            g,
            graph_height=800,
            show_node_label=True,
            show_edge_label=False,
            edge_curvature=0.1,
            use_node_size_normalization=True,
            node_size_normalization_min=10,
            node_size_normalization_max=50,
            node_drag_fix=True,
            node_hover_neighborhood=True,
            use_edge_size_normalization=True,
            edge_size_normalization_max=4,
            many_body_force_strength=-3000,
            zoom_factor=1.0,
        )
        fig.export_html(schema_filepath)


def sqlite_to_schema_compact(db_filepath, schema_filepath, node_type_to_color=None):
    if node_type_to_color is None:
        node_type_to_color = {}

    with sqlite3.connect(db_filepath) as conn:
        cursor = conn.cursor()

        # Nodes
        query = """
            SELECT
                type, COUNT(*) as count
            FROM
                nodes
            GROUP BY
                type
            ORDER BY
                count DESC,
                type ASC
        """
        cursor.execute(query)
        rows = cursor.fetchall()
        node_data = list(rows)

        # Edges
        query = """
            SELECT
                source_node.type AS source_type,
                target_node.type AS target_type,
                COUNT(*) AS duple_count,
                COUNT(DISTINCT edges.type) AS distinct_edge_types_count
            FROM
                edges
            JOIN
                nodes AS source_node
            ON
                edges.source_id = source_node.id
            JOIN
                nodes AS target_node
            ON
                edges.target_id = target_node.id
            GROUP BY
                source_type, target_type
            ORDER BY
                duple_count DESC
        """
        cursor.execute(query)
        rows = cursor.fetchall()
        edge_data = list(rows)

        # Graph
        g = nx.DiGraph()
        for node_type, cnt_instances in node_data:
            size = cnt_instances
            color = node_type_to_color.get(node_type, None)
            hover = (
                f"Node type: {node_type}\n\n" f"Number of instances: {cnt_instances}"
            )
            g.add_node(
                node_type, size=size, color=color, label_color=color, hover=hover
            )

        for source_type, target_type, cnt_instances, cnt_types in edge_data:
            size = cnt_instances
            color = node_type_to_color.get(source_type, None)
            hover = (
                f"Source: {source_type}\n"
                f"Target: {target_type}\n\n"
                f"Number of instances: {cnt_instances}\n"
                f"Number of represented edge types: {cnt_types}"
            )
            g.add_edge(source_type, target_type, size=size, color=color, hover=hover)

        # Plot
        fig = gv.d3(
            g,
            graph_height=800,
            show_node_label=True,
            show_edge_label=False,
            edge_curvature=0.1,
            use_node_size_normalization=True,
            node_size_normalization_min=10,
            node_size_normalization_max=50,
            node_drag_fix=True,
            node_hover_neighborhood=True,
            use_edge_size_normalization=True,
            edge_size_normalization_max=4,
            many_body_force_strength=-3000,
            zoom_factor=1.0,
        )
        fig.export_html(schema_filepath)


def sqlite_to_sql(db_filepath, sql_filepath, batch_size=10_000):
    # https://docs.python.org/3/library/sqlite3.html#sqlite3.Connection.iterdump
    with sqlite3.connect(db_filepath) as conn:
        with open(sql_filepath, "w") as f:
            for line in conn.iterdump():
                f.write(f"{line}\n")


def sqlite_to_csv(db_filepath, csv_filepath, table_name, batch_size=10_000):
    if table_name not in ("nodes", "edges"):
        message = f"Unknown table name: {table_name}"
        raise ValueError(message)

    with sqlite3.connect(db_filepath) as conn:
        with open(csv_filepath, "w", newline="") as f:
            cursor = conn.cursor()
            writer = csv.writer(f, dialect="excel", quoting=csv.QUOTE_ALL)
            # Header
            query = f"SELECT * FROM {table_name} LIMIT 0"
            cursor.execute(query)
            columns = [desc[0] for desc in cursor.description]
            writer.writerow(columns)
            # Data
            query = f"SELECT * FROM {table_name}"
            cursor.execute(query)
            while True:
                rows = cursor.fetchmany(batch_size)
                if not rows:
                    break
                writer.writerows(rows)


def sqlite_to_jsonl(db_filepath, jsonl_filepath, table_name, batch_size=10_000):
    if table_name == "nodes":
        query = """
            SELECT
                json_object(
                    'id', id,
                    'type', type,
                    'properties', json(properties)
                )
            FROM
                nodes
        """
        with sqlite3.connect(db_filepath) as conn:
            with open(jsonl_filepath, "w") as f:
                cursor = conn.cursor()
                cursor.execute(query)
                while True:
                    rows = cursor.fetchmany(batch_size)
                    if not rows:
                        break
                    f.write("\n".join(r[0] for r in rows))
                    f.write("\n")
    elif table_name == "edges":
        query = """
            SELECT
                json_object(
                    'source_id', source_id,
                    'target_id', target_id,
                    'type', type,
                    'properties', json(properties)
                )
            FROM
                edges
        """
        with sqlite3.connect(db_filepath) as conn:
            with open(jsonl_filepath, "w") as f:
                cursor = conn.cursor()
                cursor.execute(query)
                while True:
                    rows = cursor.fetchmany(batch_size)
                    if not rows:
                        break
                    f.write("\n".join(r[0] for r in rows))
                    f.write("\n")
    else:
        message = f"Unknown table name: {table_name}"
        raise ValueError(message)


def sqlite_to_graphml(db_filepath, graphml_filepath, batch_size=10_000):
    # http://graphml.graphdrawing.org
    def clean_id(item):
        return item.replace("\\", '\\\\"').replace('"', '\\"')

    def clean_val(item):
        if isinstance(item, (bool, int, float)):
            return item
        else:
            item_str = orjson.dumps(item).decode("utf-8")
            item_xml = saxutils.escape(item_str)
            return item_xml

    with sqlite3.connect(db_filepath) as conn:
        with open(graphml_filepath, "w") as f:
            cursor = conn.cursor()

            # Head
            head = """<?xml version="1.0" encoding="UTF-8"?>
<graphml xmlns="http://graphml.graphdrawing.org/xmlns"
         xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
         xsi:schemaLocation="http://graphml.graphdrawing.org/xmlns
                             http://graphml.graphdrawing.org/xmlns/1.0/graphml.xsd">
"""
            f.write(head)

            # Attributes
            def attr_val_to_type(val):
                attr_type = "string"
                if isinstance(val, bool):
                    attr_type = "boolean"
                elif isinstance(val, int):
                    attr_type = "long"  # alternative: "int"
                elif isinstance(val, float):
                    attr_type = "double"  # alternative: "float"
                return attr_type

            attributes = {}
            attr_cnt = 0
            # - Collect node attributes
            attributes["_node_type"] = {
                "id": f"d{attr_cnt}",
                "for": "node",
                "name": "node_type",
                "type": "string",
            }
            attr_cnt += 1
            query = "SELECT properties FROM nodes"
            cursor.execute(query)
            while True:
                rows = cursor.fetchmany(batch_size)
                if not rows:
                    break
                result = []
                for (properties_str,) in rows:
                    properties = orjson.loads(properties_str)
                    for name, val in properties.items():
                        key = f"node_{name}"
                        if key not in attributes:
                            attributes[key] = {
                                "id": f"d{attr_cnt}",
                                "for": "node",
                                "name": name,
                                "type": attr_val_to_type(val),
                            }
                            attr_cnt += 1
            # - Collect edge attributes
            attributes["_edge_type"] = {
                "id": f"d{attr_cnt}",
                "for": "edge",
                "name": "edge_type",
                "type": "string",
            }
            attr_cnt += 1
            query = "SELECT properties FROM edges"
            cursor.execute(query)
            while True:
                rows = cursor.fetchmany(batch_size)
                if not rows:
                    break
                result = []
                for (properties_str,) in rows:
                    properties = orjson.loads(properties_str)
                    for name, val in properties.items():
                        key = f"edge_{name}"
                        if key not in attributes:
                            attributes[key] = {
                                "id": f"d{attr_cnt}",
                                "for": "edge",
                                "name": name,
                                "type": attr_val_to_type(val),
                            }
                            attr_cnt += 1
            # - Write both
            lines = []
            for _, attr in attributes.items():
                lines.append(
                    "<"
                    f'key id="{attr["id"]}" '
                    f'for="{attr["for"]}" '
                    f'attr.name="{attr["name"]}" '
                    f'attr.type="{attr["type"]}"'
                    "/>"
                )
            f.write("\n".join(lines))

            # Graph
            f.write('\n<graph id="knowledge_graph" edgedefault="directed">\n')

            # Nodes
            query = "SELECT * FROM nodes"
            cursor.execute(query)
            while True:
                rows = cursor.fetchmany(batch_size)
                if not rows:
                    break
                result = []
                for node_id, node_type, node_properties in rows:
                    lines = []
                    lines.append(f'<node id="{clean_id(node_id)}">')
                    attr = {"_node_type": node_type}
                    for k, v in orjson.loads(node_properties).items():
                        key = f"node_{k}"
                        attr[key] = v
                    for key, val in attr.items():
                        lines.append(
                            f' <data key="{attributes[key]["id"]}">{clean_val(val)}</data>'
                        )
                    lines.append(" </node>")
                    result.extend(lines)
                f.write("\n".join(result))
                f.write("\n")

            # Edges
            query = "SELECT * FROM edges"
            cursor.execute(query)
            edge_id = 0
            while True:
                rows = cursor.fetchmany(batch_size)
                if not rows:
                    break
                result = []
                for source_id, target_id, edge_type, edge_properties in rows:
                    lines = []
                    lines.append(
                        f'<edge id="{edge_id}" '
                        f'source="{clean_id(source_id)}" '
                        f'target="{clean_id(target_id)}">'
                    )
                    attr = {"_edge_type": edge_type}
                    for k, v in orjson.loads(edge_properties).items():
                        key = f"edge_{k}"
                        attr[key] = v
                    for key, val in attr.items():
                        lines.append(
                            f' <data key="{attributes[key]["id"]}">{clean_val(val)}</data>'
                        )
                    lines.append("</edge>")
                    result.extend(lines)
                    edge_id += 1
                f.write("\n".join(result))
                f.write("\n")

            # Tail
            tail = """
</graph>
</graphml>
"""
            f.write(tail)


def clean(item):
    if isinstance(item, (int, float)):
        s = str(item)
    else:
        s = orjson.dumps(item).decode("utf-8")
        if not s.startswith('"'):
            s = orjson.dumps(s).decode("utf-8")
    return s


def sqlite_to_metta_repr1(db_filepath, metta_filepath, batch_size=10_000):
    with sqlite3.connect(db_filepath) as conn:
        with open(metta_filepath, "w") as f:
            cursor = conn.cursor()

            # Types
            f.write("; Types\n")
            # Node types
            f.write("(: NodeType Type)\n")
            query = "SELECT DISTINCT type FROM nodes"
            cursor.execute(query)
            while True:
                rows = cursor.fetchmany(batch_size)
                if not rows:
                    break
                result = []
                for (node_type,) in rows:
                    # (: node_type NodeType)
                    node_type = clean(node_type)
                    line = f"(: {node_type} NodeType)"
                    result.append(line)
                f.write("\n".join(result))
                f.write("\n")

            # Nodes
            f.write("; Nodes\n")
            query = "SELECT * FROM nodes"
            cursor.execute(query)
            while True:
                rows = cursor.fetchmany(batch_size)
                if not rows:
                    break
                result = []
                for node_id, node_type, _ in rows:
                    node_id = clean(node_id)
                    node_type = clean(node_type)
                    # node type definitions:
                    #   (: node_id node_type)
                    line = f"(: {node_id} {node_type})"
                    result.append(line)
                f.write("\n".join(result))
                f.write("\n")

            # Edges
            f.write("; Edges\n")
            query = "SELECT * FROM edges"
            cursor.execute(query)
            while True:
                rows = cursor.fetchmany(batch_size)
                if not rows:
                    break
                result = []
                for source_id, target_id, edge_type, _ in rows:
                    source_id = clean(source_id)
                    target_id = clean(target_id)
                    edge_type = clean(edge_type)
                    line = f"({source_id} {edge_type} {target_id})"
                    result.append(line)
                f.write("\n".join(result))
                f.write("\n")


def sqlite_to_metta_repr2(db_filepath, metta_filepath, batch_size=10_000):
    property_relation = '"has_property"'

    with sqlite3.connect(db_filepath) as conn:
        with open(metta_filepath, "w") as f:
            cursor = conn.cursor()

            # Types
            f.write("; Types\n")
            f.write("(: NodeType Type)\n")
            f.write("(: EdgeType Type)\n")
            # 1. Node types
            query = "SELECT DISTINCT type FROM nodes"
            cursor.execute(query)
            while True:
                rows = cursor.fetchmany(batch_size)
                if not rows:
                    break
                result = []
                for (node_type,) in rows:
                    # (: node_type NodeType)
                    node_type = clean(node_type)
                    line = f"(: {node_type} NodeType)"
                    result.append(line)
                f.write("\n".join(result))
                f.write("\n")
            # 2. Edge types
            query = "SELECT DISTINCT type FROM edges"
            cursor.execute(query)
            while True:
                rows = cursor.fetchmany(batch_size)
                if not rows:
                    break
                result = []
                for (edge_type,) in rows:
                    # (: edge_type Edge)
                    edge_type = clean(edge_type)
                    line = f"(: {edge_type} EdgeType)"
                    result.append(line)
                f.write("\n".join(result))
                f.write("\n")

            # Nodes, their types, and their properties
            f.write("; Nodes\n")
            query = "SELECT * FROM nodes"
            cursor.execute(query)
            while True:
                rows = cursor.fetchmany(batch_size)
                if not rows:
                    break
                result = []
                for node_id, node_type, node_properties_str in rows:
                    node_id = clean(node_id)
                    node_type = clean(node_type)
                    node_properties = orjson.loads(node_properties_str)
                    # node type definitions:
                    #   (: node_id node_type)
                    line = f"(: {node_id} {node_type})"
                    result.append(line)
                    # node properties:
                    #   (has_property node_id (node_property_key node_property_value))
                    for key, val in node_properties.items():
                        key = clean(key)
                        val = clean(val)
                        line = f"({property_relation} {node_id} ({key} {val}))"
                        result.append(line)
                f.write("\n".join(result))
                f.write("\n")

            # Edges, their types, and their properties
            f.write("; Edges\n")
            query = "SELECT * FROM edges"
            cursor.execute(query)
            while True:
                rows = cursor.fetchmany(batch_size)
                if not rows:
                    break
                result = []
                for i, (
                    source_id,
                    target_id,
                    edge_type,
                    edge_properties_str,
                ) in enumerate(rows):
                    edge_id = f'"e{i}"'
                    source_id = clean(source_id)
                    target_id = clean(target_id)
                    edge_type = clean(edge_type)
                    edge_properties = orjson.loads(edge_properties_str)
                    # edge type definitions:
                    #   (: edge_id edge_type)
                    line = f"(: {edge_id} {edge_type})"
                    result.append(line)
                    # edges in prefix notation:
                    #   (edge_id (predicate subject object))
                    line = f"({edge_id} ({edge_type} {source_id} {target_id}))"
                    result.append(line)

                    # edge properties:
                    #   (has_property edge_id (edge_property_key edge_property_value))
                    for key, val in edge_properties.items():
                        key = clean(key)
                        val = clean(val)
                        line = f"({property_relation} {edge_id} ({key} {val}))"
                        result.append(line)
                f.write("\n".join(result))
                f.write("\n")


def sqlite_to_metta_repr3(db_filepath, metta_filepath, batch_size=10_000):
    def dict_generator(d, pre=None):
        pre = pre[:] if pre else []
        if isinstance(d, dict):
            for key, value in d.items():
                if isinstance(value, dict):
                    yield from dict_generator(value, pre + [key])
                elif isinstance(value, list) or isinstance(value, tuple):
                    for k0, v0 in enumerate(value):
                        if isinstance(v0, list) or isinstance(v0, tuple):
                            for k1, v1 in enumerate(v0):
                                if isinstance(v1, list) or isinstance(v1, tuple):
                                    for k2, v2 in enumerate(v1):
                                        if isinstance(v2, list) or isinstance(
                                            v2, tuple
                                        ):
                                            for k3, v3 in enumerate(v2):
                                                if isinstance(v3, list) or isinstance(
                                                    v3, tuple
                                                ):
                                                    for k4, v4 in enumerate(v3):
                                                        if isinstance(
                                                            v4, list
                                                        ) or isinstance(v4, tuple):
                                                            raise NotImplementedError()
                                                        else:
                                                            yield from dict_generator(
                                                                v4,
                                                                pre
                                                                + [
                                                                    (
                                                                        key,
                                                                        k0,
                                                                        k1,
                                                                        k2,
                                                                        k3,
                                                                        k4,
                                                                    )
                                                                ],
                                                            )
                                                else:
                                                    yield from dict_generator(
                                                        v3,
                                                        pre + [(key, k0, k1, k2, k3)],
                                                    )
                                        else:
                                            yield from dict_generator(
                                                v2, pre + [(key, k0, k1, k2)]
                                            )
                                else:
                                    yield from dict_generator(v1, pre + [(key, k0, k1)])
                        else:
                            yield from dict_generator(v0, pre + [(key, k0)])
                else:
                    yield pre + [key, value]
        else:
            yield pre + [d]

    def dict_to_metta(data):
        output = []
        for path in dict_generator(data):
            s = path[-1]
            if isinstance(s, str):
                s = clean(s)
            for item in reversed(path[:-1]):
                if isinstance(item, tuple):
                    item_str = " ".join(clean(x) for x in item)
                    s = f"({item_str} {s})"
                else:
                    s = f"({clean(item)} {s})"
            output.append(s)
        return output

    with sqlite3.connect(db_filepath) as conn:
        with open(metta_filepath, "w") as f:
            cursor = conn.cursor()
            cnt = 0

            # Nodes
            query = "SELECT * FROM nodes"
            cursor.execute(query)
            while True:
                rows = cursor.fetchmany(batch_size)
                if not rows:
                    break
                result = []
                for node_id, node_type, node_properties_str in rows:
                    node_properties = orjson.loads(node_properties_str)
                    item = {
                        cnt: {
                            "id": node_id,
                            "type": node_type,
                            "properties": node_properties,
                        }
                    }
                    metta_lines = dict_to_metta(item)
                    result.extend(metta_lines)
                    cnt += 1
                f.write("\n".join(result))
                f.write("\n")

            # Edges
            query = "SELECT * FROM edges"
            cursor.execute(query)
            while True:
                rows = cursor.fetchmany(batch_size)
                if not rows:
                    break
                result = []
                for source_id, target_id, edge_type, edge_properties_str in rows:
                    edge_properties = orjson.loads(edge_properties_str)
                    item = {
                        cnt: {
                            "source_id": source_id,
                            "target_id": target_id,
                            "type": edge_type,
                            "properties": edge_properties,
                        }
                    }
                    metta_lines = dict_to_metta(item)
                    result.extend(metta_lines)
                    cnt += 1
                f.write("\n".join(result))
                f.write("\n")



import hashlib
import math
import os
import re
import subprocess
import tarfile
import time
import zipfile

import requests
from bs4 import BeautifulSoup


# Basic OS functionality


def run_shell_command(command):
    result = subprocess.run(
        command, check=True, capture_output=True, text=True, shell=True
    )
    return result.stdout


def create_dir(dirpath):
    os.makedirs(dirpath, exist_ok=True)


def delete_file(filepath):
    try:
        os.remove(filepath)
    except OSError:
        pass


def get_file_size(filepath):
    try:
        size = os.path.getsize(filepath)
    except FileNotFoundError:
        size = 0
    return size


def extract_tar_gz(filepath):
    directory = os.path.dirname(filepath)
    with tarfile.open(filepath, "r:gz") as tar:
        tar.extractall(path=directory)


def extract_zip(filepath):
    directory = os.path.dirname(filepath)
    with zipfile.ZipFile(filepath, "r") as zip_ref:
        zip_ref.extractall(path=directory)


# Basic web functionality


def get_request_with_retries(url, num_retries=3, delay_in_sec=15, **kwargs):
    if "timeout" not in kwargs:
        kwargs["timeout"] = 5

    for _ in range(num_retries + 1):
        try:
            response = requests.get(url, **kwargs)
            response.raise_for_status()
            break
        except requests.exceptions.RequestException as e:
            last_exception = e
        time.sleep(delay_in_sec)
    else:
        raise last_exception
    return response


def has_internet_connection():
    urls = ["https://1.1.1.1", "https://8.8.8.8"]
    for url in urls:
        response = get_request_with_retries(url, num_retries=2, delay_in_sec=10)
        if response.status_code == 200:
            return True
    return False


def ensure_internet_connection():
    if not has_internet_connection():
        raise Exception("No internet connection.")


def get_remote_file_size(url):
    ensure_internet_connection()

    try:
        # Attempt 1: HEAD request with "content-length" response attribute
        headers = {
            "User-Agent": "Mozilla/5.0 (X11; Linux x86_64; rv:127.0) "
            "Gecko/20100101 Firefox/127.0"
        }
        response = requests.head(url, headers=headers, allow_redirects=True, timeout=10)
        remote_size = int(response.headers.get("content-length", 0))

        # Attempt 2: GET request with Range header and "content-range" response attribute
        if remote_size == 0:
            headers = {
                "User-Agent": "Mozilla/5.0 (X11; Linux x86_64; rv:127.0) "
                "Gecko/20100101 Firefox/127.0",
                "Range": "bytes=0-1",
            }
            response = requests.get(
                url, headers=headers, allow_redirects=True, timeout=10
            )
            content_range = response.headers.get("content-range")
            try:
                remote_size = int(content_range.split("/")[-1])
            except Exception:
                return 0
        return remote_size
    except requests.exceptions.RequestException as e:
        print(f"Error fetching remote size: {str(e)}")
        return 0


def download_file(url, filepath, local_size=0):
    headers = {"Range": f"bytes={local_size}-"}
    response = requests.get(url, headers=headers, stream=True)
    with open(filepath, "ab") as f:
        for data in response.iter_content(chunk_size=1024):
            f.write(data)


def fetch_file(url, filepath):
    remote_size = get_remote_file_size(url)
    local_size = get_file_size(filepath)
    if local_size == 0:
        # Found no local copy of the file. Starting the download.
        download_file(url, filepath, local_size)
    elif remote_size == 0:
        # Found a local copy but couldn't determine the remote file size.
        # Starting a fresh download to overwrite the possibly partial local file.
        delete_file(filepath)
        local_size = 0
        download_file(url, filepath, local_size)
    else:
        if remote_size == local_size:
            # Found a full local copy of the file
            pass
        elif remote_size > local_size:
            # Found a partial local copy of the file. Resuming the download.
            download_file(url, filepath, local_size)
        else:
            # Found a local copy but it is larger than the remote file.
            # Starting a fresh download to overwrite the local file.
            delete_file(filepath)
            local_size = 0
            download_file(url, filepath, local_size)


def is_valid_file_by_md5(filepath, md5_hash):
    with open(filepath, "rb") as f:
        bytes = f.read()
    md5_hash_calc = hashlib.md5(bytes).hexdigest()
    return md5_hash == md5_hash_calc


def is_valid_file_by_sha256(filepath, sha256_hash):
    with open(filepath, "rb") as f:
        bytes = f.read()
    sha256_hash_calc = hashlib.sha256(bytes).hexdigest()
    return sha256_hash == sha256_hash_calc


# Access to data repositories on web


def get_versions_from_figshare(dataset_id):
    url = f"https://api.figshare.com/v2/articles/{dataset_id}/versions"
    response = get_request_with_retries(url)
    raw_data = response.json()
    versions = [str(entry["version"]) for entry in raw_data]
    return versions


def get_metadata_from_figshare(dataset_id, version):
    url = f"https://api.figshare.com/v2/articles/{dataset_id}/versions/{version}"
    response = get_request_with_retries(url)
    raw_data = response.json()
    data = {}
    date = raw_data["created_date"]
    for entry in raw_data["files"]:
        name = entry["name"]
        url = entry["download_url"]
        size = entry["size"]
        md5 = entry["computed_md5"]
        data[name] = dict(version=version, date=date, url=url, md5=md5, size=size)
    return data


def get_versions_from_mendeley(dataset_id):
    url = f"https://data.mendeley.com/public-api/datasets/{dataset_id}/versions"
    response = get_request_with_retries(url)
    raw_data = response.json()
    versions = [str(entry["version"]) for entry in raw_data]
    return versions


def get_metadata_from_mendeley(dataset_id, version):
    url = (
        "https://data.mendeley.com/public-api/datasets/"
        f"{dataset_id}/files?folder_id=root&version={version}"
    )
    response = get_request_with_retries(url)
    raw_data = response.json()
    data = {}
    for entry in raw_data:
        name = entry["filename"]
        date = entry["content_details"]["created_date"]
        url = entry["content_details"]["download_url"]
        size = entry["size"]
        sha256 = entry["content_details"]["sha256_hash"]
        data[name] = dict(version=version, date=date, url=url, size=size, sha256=sha256)
    return data


def get_versions_from_monarch():
    url = "https://data.monarchinitiative.org/monarch-kg/index.html"
    response = get_request_with_retries(url)
    soup = BeautifulSoup(response.content, "html.parser")
    links = soup.find_all("a")
    date_pattern = re.compile(r"^\d{4}-\d{2}-\d{2}$")
    url_pattern = re.compile(
        r"^https://data.monarchinitiative.org/monarch-kg/\d{4}-\d{2}-\d{2}/index.html$"
    )
    versions = []
    for link in links:
        href = link.get("href")
        if href:
            text = str(link.get_text(strip=True))
            if date_pattern.match(text) and url_pattern.match(href) and text in href:
                versions.append(text)
    return versions


def get_metadata_from_monarch(version):
    url = f"https://data.monarchinitiative.org/monarch-kg/{version}/index.html"
    response = get_request_with_retries(url)
    soup = BeautifulSoup(response.content, "html.parser")
    links = soup.find_all("a")
    data = {}
    for link in links:
        href = link.get("href")
        if href:
            text = link.get_text(strip=True)
            if "." in text and ".." not in text:
                name = text
                data[name] = dict(version=version, date=version, url=href)
    return data


def get_metadata_from_primekg():
    # Caution: Uses a custom mirror because Harvard Dataverse does not enable
    # reliable programmatic access
    url = (
        "https://raw.githubusercontent.com/robert-haas/"
        "primekg-mirror/main/metadata.json"
    )
    response = get_request_with_retries(url)
    data = response.json()
    return data


# Filtering


def is_informative_value(value):
    """Find out if a value is not None, not an empty string, and not NaN."""
    if value is None or value == "":
        return False

    try:
        # Check if value is NaN
        return not math.isnan(value)
    except TypeError:
        # Value is not a float or does not support NaN, assume valid
        return True



import orjson


def read_json_file(filepath):
    with open(filepath, "rb") as f:
        data = orjson.loads(f.read())
    return data


def create_sql_schema(cursor):
    query = """
    CREATE TABLE IF NOT EXISTS nodes (
        id TEXT PRIMARY KEY,
        type TEXT NOT NULL,
        properties TEXT
    );

    CREATE TABLE IF NOT EXISTS edges (
        source_id TEXT NOT NULL,
        target_id TEXT NOT NULL,
        type TEXT NOT NULL,
        properties TEXT,
        FOREIGN KEY (source_id) REFERENCES nodes (id),
        FOREIGN KEY (target_id) REFERENCES nodes (id)
    );

    CREATE INDEX IF NOT EXISTS idx_edges_source ON edges (source_id);
    CREATE INDEX IF NOT EXISTS idx_edges_target ON edges (target_id);
"""
    cursor.executescript(query)



import csv
import os
import sqlite3

import luigi
import orjson

from .. import _shared
from .._shared.extract import is_informative_value


class MetadataFetcher:
    dataset_id = 23553114
    _cached_values = {}

    @classmethod
    def get_versions(cls):
        # Memorize
        key = "versions"
        if key not in cls._cached_values:
            cls._cached_values[key] = _shared.extract.get_versions_from_figshare(
                cls.dataset_id
            )

        # Lookup
        versions = cls._cached_values[key]
        return versions

    @classmethod
    def get_metadata(cls, version):
        # Check
        versions = cls.get_versions()
        if version not in versions:
            raise ValueError(
                f'Version "{version}" is not valid.\nAvailable options: {versions}'
            )

        # Memorize
        key = f"metadata_{version}"
        if key not in cls._cached_values:
            cls._cached_values[key] = _shared.extract.get_metadata_from_figshare(
                cls.dataset_id, version
            )

        # Lookup
        metadata = cls._cached_values[key]
        return metadata


class FetchKnowledgeGraphFile(_shared.tasks.ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "downloads")
        self.metadata = MetadataFetcher.get_metadata(self.version)

    def _detect_filename(self):
        # The filename is slightly different for different versions, hence it has to be detected
        for candidate in self.metadata:
            if candidate.startswith("OREGANO") and candidate.endswith(".tsv"):
                return candidate
        raise ValueError(
            "Could not identify the TSV file that contains Oregano's knowledge graph."
        )

    def requires(self):
        filename = self._detect_filename()
        url = self.metadata[filename]["url"]
        md5 = self.metadata[filename]["md5"]
        return _shared.tasks.DownloadFile(self.subdirpath, filename, url, md5=md5)

    def output(self):
        return self.input()


class FetchAnnotationFiles(_shared.tasks.ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "downloads")
        self.metadata = MetadataFetcher.get_metadata(self.version)
        self.filenames = [
            "ACTIVITY.tsv",
            "COMPOUND.tsv",
            "DISEASES.tsv",
            "EFFECT.tsv",
            "GENES.tsv",
            "INDICATION.tsv",
            "PATHWAYS.tsv",
            "PHENOTYPES.tsv",
            "SIDE_EFFECT.tsv",
            "TARGET.tsv",
        ]

    def requires(self):
        tasks = []
        for filename in self.filenames:
            url = self.metadata[filename]["url"]
            md5 = self.metadata[filename]["md5"]
            task = _shared.tasks.DownloadFile(self.subdirpath, filename, url, md5=md5)
            tasks.append(task)
        return tasks

    def output(self):
        return self.input()


class CreateSqliteFile(_shared.tasks.ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()
    batch_size = luigi.IntParameter(default=10_000, significant=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "results")
        self.filepath = os.path.join(self.subdirpath, "kg.sqlite")

    def requires(self):
        tasks = {
            "subdirpath": _shared.tasks.CreateDirectory(self.subdirpath),
            "kg_file": FetchKnowledgeGraphFile(self.dirpath, self.version),
            "annotation_files": FetchAnnotationFiles(self.dirpath, self.version),
        }
        return tasks

    def output(self):
        return luigi.LocalTarget(self.filepath)

    def run(self):
        with _shared.tasks.temp_output_path(self.filepath) as sqlite_filepath:
            with sqlite3.connect(sqlite_filepath) as conn:
                cursor = conn.cursor()
                self._prepare_annotations()
                _shared.transform.create_sql_schema(cursor)
                self._insert_nodes(conn, cursor)
                self._insert_edges(conn, cursor)
                conn.commit()

    @staticmethod
    def node_name_to_type(node_name):
        try:
            # Most node names contain both the type and id of a node separated
            # by a ":" character.
            node_type, _ = node_name.split(":", 1)
            node_type = node_type.lower()
        except Exception:
            # Node names that occur in "has_code" relations as target do not
            # have a ":" and explicit type.
            # Here the type "code" is assigned to such nodes to be consistent.
            node_type = "code"
        return node_type

    def _prepare_annotations(self):
        targets = self.input()["annotation_files"]

        def strip_if_str(val):
            if isinstance(val, str):
                val = val.strip()
            return val

        node_name_to_annotation_map = {}
        for target in targets:
            filepath = target.path
            with open(filepath, newline="") as f:
                reader = csv.reader(f, delimiter="\t")
                columns = next(reader)
                data_columns = columns[1:]
                for row in reader:
                    node_name = row[0]
                    annotation = {}
                    for i, col in enumerate(data_columns, 1):
                        value = strip_if_str(row[i])
                        if is_informative_value(value):
                            annotation[strip_if_str(col)] = value
                    node_name_to_annotation_map[node_name] = annotation
        self.node_name_to_annotation_map = node_name_to_annotation_map

    def _insert_nodes(self, conn, cursor):
        kg_filepath = self.input()["kg_file"].path
        sql_cmd = """
            INSERT INTO nodes (id, type, properties)
            VALUES (?, ?, ?)
        """
        with open(kg_filepath, "r", newline="") as csvfile:
            reader = csv.reader(csvfile, delimiter="\t")
            seen_nodes = set()
            batch = []
            for row in reader:
                subject, _, object = row
                for node_name in (subject, object):
                    if node_name not in seen_nodes:  # skip duplicates
                        seen_nodes.add(node_name)
                        node_id = node_name
                        node_type = self.node_name_to_type(node_name)
                        node_properties = self.node_name_to_annotation_map.get(
                            node_id, {}
                        )
                        node_properties_str = orjson.dumps(node_properties).decode(
                            "utf-8"
                        )
                        node = (node_id, node_type, node_properties_str)
                        batch.append(node)
                if len(batch) >= self.batch_size:
                    cursor.executemany(sql_cmd, batch)
                    batch = []
            if batch:
                cursor.executemany(sql_cmd, batch)
        conn.commit()

    def _insert_edges(self, conn, cursor):
        # Special aspects in handling the raw data
        # 1) This code skips redundant edges, i.e. repeated (subject, predicte, object) triples,
        #    because they do not differ in any other way as well.
        # 2) There are no edge properties in this knowledge graph
        kg_filepath = self.input()["kg_file"].path
        sql_cmd = """
            INSERT INTO edges (source_id, target_id, type, properties)
            VALUES (?, ?, ?, ?)
        """
        with open(kg_filepath, "r", newline="") as csvfile:
            reader = csv.reader(csvfile, delimiter="\t")
            seen_triples = set()
            batch = []
            for row in reader:
                subject, predicate, object = row
                source_id = subject
                target_id = object
                edge_type = predicate
                edge_properties = {}  # no edge properties in raw data
                edge_properties_str = orjson.dumps(edge_properties).decode("utf-8")
                triple = (source_id, edge_type, target_id)
                if triple not in seen_triples:  # skip duplicates
                    seen_triples.add(triple)
                    edge = (source_id, target_id, edge_type, edge_properties_str)
                    batch.append(edge)
                    if len(batch) >= self.batch_size:
                        cursor.executemany(sql_cmd, batch)
                        batch = []
            if batch:
                cursor.executemany(sql_cmd, batch)
        conn.commit()


class Oregano(_shared.base.Project):
    """Oregano Knowledge Graph.

    References
    ----------
    - Publication: https://doi.org/10.1038/s41597-023-02757-0
    - Code: https://gitub.u-bordeaux.fr/erias/oregano
    - Data: https://doi.org/10.6084/m9.figshare.23553114

    """

    _label = "oregano"
    _MetadataFetcher = MetadataFetcher
    _CreateSqliteFile = CreateSqliteFile
    _node_type_to_color = {
        "compound": "green",
        "molecule": "green",
        "gene": "blue",
        "protein": "blue",
        "disease": "red",
        "pathway": "red",
    }



import logging
import multiprocessing
import os
import sys
import warnings

import luigi

from . import tasks


class Project:
    def __init__(self, version, workdir):
        """Initialize a project instance so that tasks can be defined on it.

        Parameters
        ----------
        version : `str`
            Version of the dataset that will be downloaded and processed.
            The method :meth:`get_versions` returns all currently available
            versions.
        workdir : `str`
            Path of the working directory in which a unique subdirectory will
            be created to hold all downloaded and generated files for
            this project and version.

        Raises
        ------
        ValueError
            Raised if `version` is invalid or unavailable.
        TypeError
            Raised if `workdir` is not a string.

        Notes
        -----
        This class does not automatically download or process any data.
        Such tasks first need to be specified by calling the relevant methods
        on the project object and then passing it to the function :func:`~kgw.run`
        that builds and executes a corresponding workflow.

        """
        # Argument processing
        version = str(version)
        _available_versions = self.get_versions()
        if version.lower() == "latest":
            version = _available_versions[-1]
        if version not in _available_versions:
            msg = f'Version "{version}" is not valid.\nAvailable options: {_available_versions}'
            raise ValueError(msg)

        if not isinstance(workdir, str):
            msg = 'Argument "workdir" needs to be a string.'
            raise TypeError(msg)

        # Set instance variables
        self.version = version
        self.dirpath = os.path.join(workdir, f"{self._label}_v{version}")
        self._tasks = []

    @classmethod
    def get_versions(cls):
        """Fetch all currently available versions from the data repository of the project."""
        return cls._MetadataFetcher.get_versions()

    def to_sqlite(self):
        """Convert the knowledge graph to a file-based SQLite database.

        Generates the output file `kg.sqlite`. This database contains
        a unified representation for each knowledge graph, with the
        same schema being used for each projects. From this intermediate
        format it is possible to generate all other files, using just
        one method per output format rather than writing a custom
        converter for each project.

        References
        ----------
        - `SQLite <https://www.sqlite.org>`__
        - `Wikipedia: SQLite <https://en.wikipedia.org/wiki/SQLite>`__

        """
        tasks = [self._CreateSqliteFile(self.dirpath, self.version)]
        self._tasks.extend(tasks)

    def to_statistics(self):
        """Determine some statistical properties of the knowledge graph.

        Generates the output file `statistics.json`. This is a JSON
        file that contains data about basic statistics of the elements
        in the knowledge graph, such as node, edge and type counts.

        """
        required_tasks = [
            tasks.CreateStatisticsFile(
                self.dirpath, self.version, self._CreateSqliteFile
            ),
        ]
        self._tasks.extend(required_tasks)

    def to_schema(self):
        """Determine the schema of the knowledge graph.

        Generates the output file `schema.html`. This is a standalone
        HTML file with an interactive graph visualization of all
        entity types in the knowledge graph and the relationship types
        by which they are connected.

        References
        ----------
        - `Neo4j: Graph modeling guidelines
          <https://neo4j.com/docs/getting-started/data-modeling/guide-data-modeling/>`__
        - `Memgraph: Graph modeling
          <https://memgraph.com/docs/fundamentals/graph-modeling>`__

        """
        required_tasks = [
            tasks.CreateSchemaFile(
                self.dirpath,
                self.version,
                self._node_type_to_color,
                self._CreateSqliteFile,
            ),
        ]
        self._tasks.extend(required_tasks)

    def to_sql(self):
        """Convert the knowledge graph to a SQL text file.

        Generates the output file `kg.sql`. This is a text file with
        SQL commands that can be used to import the structure and content
        of the knowledge graph into relational database systems such as
        MySQL or PostgreSQL.

        References
        ----------
        - `StackOverflow: How to import an SQL file using the command line in MySQL?
          <https://stackoverflow.com/questions/17666249/how-to-import-an-sql-file-using-the-command-line-in-mysql>`__
        - `StackOverflow: Import SQL dump into PostgreSQL database
          <https://stackoverflow.com/questions/6842393/import-sql-dump-into-postgresql-database>`__

        """
        required_tasks = [
            tasks.CreateSqlFile(self.dirpath, self.version, self._CreateSqliteFile),
        ]
        self._tasks.extend(required_tasks)

    def to_csv(self):
        """Convert the knowledge graph to two CSV text files.

        Generates the output files `kg_nodes.csv` and `kg_edges.csv`.

        References
        ----------
        - `Wikipedia: CSV <https://en.wikipedia.org/wiki/Comma-separated_values>`__

        """
        required_tasks = [
            tasks.CreateCsvNodesFile(
                self.dirpath, self.version, self._CreateSqliteFile
            ),
            tasks.CreateCsvEdgesFile(
                self.dirpath, self.version, self._CreateSqliteFile
            ),
        ]
        self._tasks.extend(required_tasks)

    def to_jsonl(self):
        """Convert the knowledge graph to two JSON Lines text files.

        Generates the output files `kg_nodes.jsonl` and `kg_edges.jsonl`.

        References
        ----------
        - `JSONL <https://jsonlines.org>`__

        """
        required_tasks = [
            tasks.CreateJsonlNodesFile(
                self.dirpath, self.version, self._CreateSqliteFile
            ),
            tasks.CreateJsonlEdgesFile(
                self.dirpath, self.version, self._CreateSqliteFile
            ),
        ]
        self._tasks.extend(required_tasks)

    def to_metta(self, representation="spo"):
        """Convert the knowledge graph to a MeTTa text file.

        Generates the output file `kg_spo.metta`, `kg_properties_aggregated.metta` or
        `kg_properties_expanded.metta`, depending on the chosen representation.

        Caution: These representations are still subject to experimentation
        and testing. They might change in future versions of this package.

        Parameters
        ----------
        representation : str
            The format used to represent the knowledge graph in the MeTTa language.

            Available options:

            - `"spo"`: Semantic triples of the form `("subject", "predicate", "object")`.
              If properties are present in the original knowledge graph, they are
              ignored in this representation.
            - `"properties_aggregated"`: Properties (=key-value pairs) are represented
              by putting each key on a separate line, but each value is ensured to be a
              single number or string. This means values that hold a compound data type
              like a list or dict are aggregated into one string in JSON string format.
              Text identifiers of nodes are reused to create the association with their
              properties, while text identifiers of the form "e{cnt}" are introduced for
              edges to serve the same purpose.
            - `"properties_expanded"`: Properties (=key-value pairs) are represented
              by fully expanding their keys and values onto as many lines as required.
              Numerical identifiers for nodes and edges are introduced to create the
              association between these elements and their properties.

        References
        ----------
        - `MeTTa language <https://metta-lang.dev>`__
        - `OpenCog Hyperon framework <https://hyperon.opencog.org>`__

        """
        required_tasks = [
            tasks.CreateMettaFile(
                self.dirpath, self.version, representation, self._CreateSqliteFile
            ),
        ]
        self._tasks.extend(required_tasks)

    def to_graphml(self):
        """Convert the knowledge graph to a GraphML text file.

        Generates the output file `kg.graphml`.

        References
        ----------
        - `GraphML <http://graphml.graphdrawing.org>`__
        - `Wikipedia: GraphML <https://en.wikipedia.org/wiki/GraphML>`__

        """
        required_tasks = [
            tasks.CreateGraphMlFile(self.dirpath, self.version, self._CreateSqliteFile),
        ]
        self._tasks.extend(required_tasks)


def set_logging_behavior(verbose):
    # Turn off luigi's logger
    for name in ("luigi", "luigi-interface"):
        logger = logging.getLogger(name)
        if logger.handlers:
            logger.handlers.clear()
        logger.disabled = True
        warnings.filterwarnings("ignore", module=name)

    # Set up logger for this package
    logger = logging.getLogger(__name__)
    if logger.handlers:
        logger.handlers.clear()
    if verbose:
        logger.disabled = False
        logging.basicConfig(
            level=logging.INFO,
            format="%(asctime)s  %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
            handlers=[logging.StreamHandler(sys.stdout)],
            force=True,
        )
    else:
        logger.disabled = True


def run(workflow, num_workers=None, verbose=True):
    """Execute all tasks in the provided workflow according to their dependencies.

    This function uses the package Luigi [1]_ to build
    a dependency graph of all tasks defined in the workflow and
    execute them in parallel with multiple worker processes.

    Parameters
    ----------
    workflow : `~kgw._shared.base.Project`, or list of `~kgw._shared.base.Project`
        Specification of a workflow in form of a single or multiple
        project objects. A project object provides several methods that can
        be called in order to add specific tasks.
        For example, calling the method `to_csv()` will store a task in the
        object that represents the conversion of the project's knowledge graph
        into CSV format. The workflow engine automatically detects these tasks,
        inspects their dependencies, and schedules all necessary steps in the
        correct order.
    num_workers : `int`, optional, default=4*cpu_count
        The number of worker processes to run tasks in parallel. If not
        specified, it defaults to 4 times the number of CPU cores available
        on the machine.
    verbose : `bool`, optional, default=True
        If `True`, a log of tasks and a summary of results is written to stdout.
        If `False`, no text is printed.

    Returns
    -------
    success : `bool`
        Returns `True` if all tasks were successfully scheduled and executed,
        otherwise `False`. A failed run can be resumed from an
        intermediate state without re-running previously completed tasks.

    Raises
    ------
    TypeError
        Raised if `workflow` is not a project object or list of such objects,
        or if `num_workers` is not an integer, or if `verbose` is not a boolean.
    ValueError
        Raised if `workflow` is an empty list.

    References
    ----------
    .. [1] Luigi Documentation, https://luigi.readthedocs.io

    """
    # Argument processing
    if isinstance(workflow, Project):
        workflows = [workflow]
    elif isinstance(workflow, (list, tuple, set)) and all(
        isinstance(w, Project) for w in workflow
    ):
        workflows = list(workflow)
    else:
        msg = 'Argument "workflow" needs to be a project or a list of projects.'
        raise TypeError(msg)
    if len(workflows) == 0:
        msg = "Got an empty list of workflows."
        raise ValueError(msg)

    if num_workers is None:
        num_workers = multiprocessing.cpu_count() * 4
    elif not isinstance(num_workers, int):
        msg = 'Argument "num_workers" needs to be an integer or None.'
        raise TypeError(msg)

    if not isinstance(verbose, bool):
        msg = 'Argument "verbose" needs to be True or False.'
        raise TypeError(msg)

    # Set logging behavior
    set_logging_behavior(verbose)

    # Collect tasks
    tasks = []
    for workflow in workflows:
        tasks.extend(workflow._tasks)

    # Run tasks
    if verbose:
        msg = "Log of performed tasks"
        print(msg)
        print("=" * len(msg))
        print()
    result = luigi.build(
        tasks,
        local_scheduler=True,
        detailed_summary=True,
        workers=num_workers,
    )

    # Report results
    def crop_results(text):
        prefix = "\n===== Luigi Execution Summary =====\n\n"
        postfix = "\n\n===== Luigi Execution Summary =====\n"
        if text.startswith(prefix):
            text = text[len(prefix) :]
        if text.endswith(postfix):
            text = text[: -len(postfix)]
        return text

    if verbose:
        print()
        print()
        msg = "Summary of workflow results"
        print(msg)
        print("=" * len(msg))
        print()
        print(crop_results(result.summary_text))
    return result.scheduling_succeeded



import os
import sqlite3

import luigi
import orjson

from .. import _shared
from .._shared.extract import is_informative_value


class MetadataFetcher:
    dataset_id = 22828196
    _cached_values = {}

    @classmethod
    def get_versions(cls):
        # Memorize
        key = "versions"
        if key not in cls._cached_values:
            cls._cached_values[key] = _shared.extract.get_versions_from_figshare(
                cls.dataset_id
            )

        # Lookup
        versions = cls._cached_values[key]
        return versions

    @classmethod
    def get_metadata(cls, version):
        # Check
        versions = cls.get_versions()
        if version not in versions:
            raise ValueError(
                f'Version "{version}" is not valid.\nAvailable options: {versions}'
            )

        # Memorize
        key = f"metadata_{version}"
        if key not in cls._cached_values:
            cls._cached_values[key] = _shared.extract.get_metadata_from_figshare(
                cls.dataset_id, version
            )

        # Lookup
        metadata = cls._cached_values[key]
        return metadata


class CreateSqliteFile(_shared.tasks.ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()
    batch_size = luigi.IntParameter(default=10_000, significant=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "results")
        self.filepath = os.path.join(self.subdirpath, "kg.sqlite")

    def requires(self):
        tasks = {
            "subdirpath": _shared.tasks.CreateDirectory(self.subdirpath),
            "nodes_file": FetchNodesFile(self.dirpath, self.version),
            "edges_file": FetchEdgesFile(self.dirpath, self.version),
        }
        return tasks

    def output(self):
        return luigi.LocalTarget(self.filepath)

    def run(self):
        with _shared.tasks.temp_output_path(self.filepath) as sqlite_filepath:
            with sqlite3.connect(sqlite_filepath) as conn:
                cursor = conn.cursor()
                _shared.transform.create_sql_schema(cursor)
                self._insert_nodes(conn, cursor)
                self._insert_edges(conn, cursor)
                conn.commit()

    def _insert_nodes(self, conn, cursor):
        # Special aspects in handling the raw data
        # 1) The dict keys in the raw data can be ignored, because source ids, edge types,
        #    and target ids are encoded in a cleaner way in the dict values.
        nodes_filepath = self.input()["nodes_file"].path
        data_nodes = _shared.transform.read_json_file(nodes_filepath)
        sql_cmd = """
            INSERT INTO nodes (id, type, properties)
            VALUES (?, ?, ?)
        """
        skipped_keys = ["entity", "type"]
        batch = []
        for entry in data_nodes.values():
            entry = entry[0]
            node_id = entry["entity"]
            node_type = entry["type"]
            node_properties = {
                key: val
                for key, val in entry.items()
                if key not in skipped_keys and is_informative_value(val)
            }
            node_properties_str = orjson.dumps(node_properties).decode("utf-8")
            node = (node_id, node_type, node_properties_str)
            batch.append(node)
            if len(batch) >= self.batch_size:
                cursor.executemany(sql_cmd, batch)
                batch = []
        if batch:
            cursor.executemany(sql_cmd, batch)
        conn.commit()

    def _insert_edges(self, conn, cursor):
        # Special aspects in handling the raw data
        # 1) The dict keys in the raw data can be ignored, because source ids, edge types,
        #    and target ids are encoded in a cleaner way in the dict values.
        # 2) Some dict values are mostly redundant and are therefore skipped,
        #    e.g. source and source type are part of the node data.
        edges_filepath = self.input()["edges_file"].path
        data_edges = _shared.transform.read_json_file(edges_filepath)
        sql_cmd = """
            INSERT INTO edges (source_id, target_id, type, properties)
            VALUES (?, ?, ?, ?)
        """
        skipped_keys = [
            "source entity",
            "target entity",
            "relationship",
            "source",
            "target",
            "source type",
            "target type",
        ]
        batch = []
        for entry in data_edges.values():
            source_id = entry["source entity"]
            target_id = entry["target entity"]
            edge_type = entry["relationship"]
            edge_properties = {
                key: val
                for key, val in entry.items()
                if key not in skipped_keys and is_informative_value(val)
            }
            edge_properties_str = orjson.dumps(edge_properties).decode("utf-8")
            edge = (source_id, target_id, edge_type, edge_properties_str)
            batch.append(edge)
            if len(batch) >= self.batch_size:
                cursor.executemany(sql_cmd, batch)
                batch = []
        if batch:
            cursor.executemany(sql_cmd, batch)
        conn.commit()


class FetchNodesFile(_shared.tasks.ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "downloads")
        self.metadata = MetadataFetcher.get_metadata(self.version)

    def requires(self):
        filename = "Entity_Info.json"
        url = self.metadata[filename]["url"]
        md5 = self.metadata[filename]["md5"]
        return _shared.tasks.DownloadFile(self.subdirpath, filename, url, md5=md5)

    def output(self):
        return self.input()


class FetchEdgesFile(_shared.tasks.ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "downloads")
        self.metadata = MetadataFetcher.get_metadata(self.version)

    def requires(self):
        filename = "Relation_Info.json"
        url = self.metadata[filename]["url"]
        md5 = self.metadata[filename]["md5"]
        return _shared.tasks.DownloadFile(self.subdirpath, filename, url, md5=md5)

    def output(self):
        return self.input()


class Hald(_shared.base.Project):
    """Human Aging and Longevity Dataset (HALD).

    References
    ----------
    - Publication: https://doi.org/10.1038/s41597-023-02781-0
    - Website: https://bis.zju.edu.cn/hald
    - Code: https://github.com/zexuwu/hald
    - Data: https://doi.org/10.6084/m9.figshare.22828196

    """

    _label = "hald"
    _MetadataFetcher = MetadataFetcher
    _CreateSqliteFile = CreateSqliteFile
    _node_type_to_color = {
        "Pharmaceutical Preparations": "green",
        "Toxin": "green",
        "Gene": "blue",
        "Peptide": "blue",
        "Protein": "blue",
        "RNA": "blue",
        "Disease": "red",
    }

    def to_schema(self):
        tasks = [
            _shared.tasks.CreateCompactSchemaFile(
                self.dirpath,
                self.version,
                self._node_type_to_color,
                self._CreateSqliteFile,
            ),
        ]
        self._tasks.extend(tasks)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Setup script for configuring, packaging, distributing and installing this Python package.
#
# Package information
# - Name:    kgw
# - Author:  Robert Haas
# - Email:   robert.haas@protonmail.com
# - License: See LICENSE in the package root directory
# - Copyright notice: See NOTICE in the package root directory
#
# References for best practices in creating Python packages
# - Python Packaging User Guide (PyPUG): https://packaging.python.org
# - Python Packaging Authority (PyPA):   https://www.pypa.io
# - Python Package Index (PyPI):         https://pypi.org
# - Setuptools:                          https://setuptools.readthedocs.io
# - PEP 440 about versioning:            https://www.python.org/dev/peps/pep-0440
# - Semantic Versioning (SemVer):        https://semver.org

import re
from codecs import open
from os import path

from setuptools import find_packages, setup


def locate_package_directory():
    """Identify directory of the package and its associated files."""
    try:
        return path.abspath(path.dirname(__file__))
    except Exception:
        message = (
            "The directory in which the package and its "
            "associated files are stored could not be located."
        )
        raise ValueError(message)


def read_file(filepath):
    """Read content from an UTF-8 encoded text file"""
    with open(filepath, "r", encoding="utf-8") as file_handle:
        text = file_handle.read()
    return text


def load_long_description(pkg_dir):
    """Load long description from file README.rst"""
    try:
        filepath_readme = path.join(pkg_dir, "README.rst")
        return read_file(filepath_readme)
    except Exception:
        message = "Long description could not be read from README.rst"
        raise ValueError(message)


def is_canonical(version):
    """Check if a version string is in canonical format of PEP 440."""
    # Source: https://www.python.org/dev/peps/pep-0440
    pattern = (
        r"^([1-9][0-9]*!)?(0|[1-9][0-9]*)(\.(0|[1-9][0-9]*))"
        r"*((a|b|rc)(0|[1-9][0-9]*))?(\.post(0|[1-9][0-9]*))"
        r"?(\.dev(0|[1-9][0-9]*))?$"
    )
    return re.match(pattern, version) is not None


def load_version(pkg_dir, pkg_name):
    """Load version from variable __version__ in file __init__.py"""
    try:
        # Read file
        filepath_init = path.join(pkg_dir, pkg_name, "__init__.py")
        file_content = read_file(filepath_init)
        # Parse version string with regular expression
        re_for_version = re.compile(r"""__version__\s+=\s+['"](.*)['"]""")
        match = re_for_version.search(file_content)
        version_string = match.group(1)
    except Exception:
        message = (
            "Version could not be read from variable " "__version__ in file __init__.py"
        )
        raise ValueError(message)
    # Check validity
    if not is_canonical(version_string):
        message = (
            'The detected version string "{}" is not in canonical '
            "format as defined in PEP 440.".format(version_string)
        )
        raise ValueError(message)
    return version_string


PKG_NAME = "kgw"
PKG_DIR = locate_package_directory()

setup(
    # Basic package information
    name=PKG_NAME,
    version=load_version(PKG_DIR, PKG_NAME),
    author="Robert Haas",
    author_email="robert.haas@protonmail.com",
    description="Knowledge Graph Workflows",
    long_description=load_long_description(PKG_DIR),
    url="https://github.com/robert-haas/kgw",
    license="Apache License, Version 2.0",
    # Classifiers: available ones listed at https://pypi.org/classifiers
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering",
    ],
    # Included files
    # a) auto-detected Python packages
    packages=find_packages(),
    # b) data files that are specified in the MANIFEST.in file
    include_package_data=True,
    # Dependencies that need to be fulfilled
    python_requires=">=3.6",
    # Dependencies that are downloaded by pip on installation
    install_requires=[
        "bs4",
        "gravis",
        "luigi",
        "networkx",
        "orjson",
        "requests",
    ],
    # Capability of running in compressed form
    zip_safe=False,
)



import json
import os

import pytest

import kgw

from . import utils


def test_package_api(workdir):
    # Version
    version = kgw.__version__
    assert isinstance(version, str)
    assert "." in version

    # Use the project classes with various arguments
    version = "latest"
    hald = kgw.biomedicine.Hald(version, workdir)
    oregano = kgw.biomedicine.Oregano(version, workdir)
    primekg = kgw.biomedicine.PrimeKg(version, workdir)
    monarchkg = kgw.biomedicine.MonarchKg(version, workdir)
    ckg = kgw.biomedicine.Ckg(version, workdir)
    with pytest.raises(ValueError):
        kgw.biomedicine.Hald("nonsense", workdir)
    with pytest.raises(TypeError):
        kgw.biomedicine.Hald(version, 123)

    # Fetch project versions
    for proj in [hald, oregano, primekg, monarchkg, ckg]:
        versions = proj.get_versions()
        assert isinstance(versions, list)
        for v in versions:
            assert isinstance(v, str)
            assert len(v) > 0

    # Use the run function with various argument combinations
    for workflow in [hald, [hald, oregano], (hald, oregano), set([hald, oregano])]:
        kgw.run(workflow)
        kgw.run(workflow, 3)
        kgw.run(workflow, 3, True)
        kgw.run(workflow=workflow, num_workers=42, verbose=False)

    for val in [3.14, "hi"]:
        with pytest.raises(TypeError):
            kgw.run(val)
        with pytest.raises(TypeError):
            kgw.run(hald, val)
        with pytest.raises(TypeError):
            kgw.run(hald, 2, val)
    with pytest.raises(ValueError):
        kgw.run([])


def test_projects_latest(workdir):
    # All methods of small enough projects
    project_classes = [
        kgw.biomedicine.Hald,
        kgw.biomedicine.Oregano,
        kgw.biomedicine.PrimeKg,
        kgw.biomedicine.MonarchKg,
    ]
    tasks = []
    for project_class in project_classes:
        nonsense_version = -123
        with pytest.raises(ValueError) as excinfo:
            proj = project_class(nonsense_version, workdir)
        msg = f'Version "{nonsense_version}" is not valid.\nAvailable options: '
        assert str(excinfo.value).startswith(msg)

        proj = project_class("latest", workdir)
        proj.to_sqlite()
        proj.to_statistics()
        proj.to_schema()
        proj.to_sql()
        proj.to_csv()
        proj.to_jsonl()
        proj.to_graphml()
        proj.to_metta(representation="spo")
        proj.to_metta(representation="properties_aggregated")
        proj.to_metta(representation="properties_expanded")
        tasks.append(proj)

    # Some methods of too large projects
    ckg = kgw.biomedicine.Ckg("latest", workdir)
    ckg.to_statistics()
    ckg.to_schema()
    ckg.to_metta(representation="spo")
    ckg.to_metta(representation="properties_aggregated")
    ckg.to_metta(representation="properties_expanded")
    tasks.append(ckg)

    success = kgw.run(tasks)
    assert success, "A part of the workflow failed"


def test_file_statistics(workdir, project_params):
    project_name, project_class, project_version, num_nodes, num_edges = project_params
    project = project_class(project_version, workdir)
    project.to_statistics()
    kgw.run([project])

    # File checks
    project_dir = utils.get_project_dir(project_name, project_version)
    filepath = os.path.join(workdir, project_dir, "results", "statistics.json")
    utils.check_file(filepath)

    # Content checks
    data = utils.load_json_file(filepath)
    assert isinstance(data, dict)
    for key in ("num_nodes", "num_edges", "num_node_types", "num_edge_types"):
        val = data[key]
        assert isinstance(val, int)
        assert val > 0

    for key in ("node_types", "edge_types"):
        val = data[key]
        assert isinstance(val, dict)
        assert len(val) > 0
        for typ, cnt in val.items():
            assert isinstance(typ, str)
            assert isinstance(cnt, int)
            assert len(typ) > 0
            assert cnt > 0

    # Comparisons
    assert num_nodes == data["num_nodes"]
    assert num_edges == data["num_edges"]


def test_file_schema(workdir, project_params):
    project_name, project_class, project_version, num_nodes, num_edges = project_params
    project = project_class(project_version, workdir)
    project.to_schema()
    kgw.run([project])

    # File checks
    project_dir = utils.get_project_dir(project_name, project_version)
    filepath = os.path.join(workdir, project_dir, "results", "schema.html")
    utils.check_file(filepath)


def test_file_sql(workdir, project_params):
    project_name, project_class, project_version, num_nodes, num_edges = project_params
    project = project_class(project_version, workdir)
    project.to_sql()
    kgw.run([project])

    # File checks
    project_dir = utils.get_project_dir(project_name, project_version)
    filepath = os.path.join(workdir, project_dir, "results", "kg.sql")
    utils.check_file(filepath)

    # Content checks
    text = utils.load_text_file(filepath)
    assert text.startswith("BEGIN TRANSACTION;")
    assert "CREATE TABLE nodes" in text
    assert "CREATE TABLE edges" in text
    assert 'INSERT INTO "nodes"' in text
    assert 'INSERT INTO "edges"' in text
    assert "CREATE INDEX" in text
    assert text.endswith("COMMIT;\n")


def test_file_csv(workdir, project_params):
    project_name, project_class, project_version, num_nodes, num_edges = project_params
    project = project_class(project_version, workdir)
    project.to_csv()
    kgw.run([project])

    # File checks
    project_dir = utils.get_project_dir(project_name, project_version)
    filepath_nodes = os.path.join(workdir, project_dir, "results", "kg_nodes.csv")
    filepath_edges = os.path.join(workdir, project_dir, "results", "kg_edges.csv")
    utils.check_file(filepath_nodes)
    utils.check_file(filepath_edges)

    # Content checks
    node_data, node_columns = utils.load_csv_file(filepath_nodes)
    assert node_columns == ["id", "type", "properties"]
    for row in node_data:
        nid, ntype, nprop_str = row
        assert isinstance(nid, str)
        assert isinstance(ntype, str)
        assert isinstance(nprop_str, str)
        assert len(nid) > 0
        assert len(ntype) > 0
        assert len(nprop_str) > 0
        nprop = json.loads(nprop_str)
        assert isinstance(nprop, dict)

    edge_data, edge_columns = utils.load_csv_file(filepath_edges)
    assert edge_columns == ["source_id", "target_id", "type", "properties"]
    for row in edge_data:
        source_id, target_id, etype, eprop_str = row
        assert isinstance(source_id, str)
        assert isinstance(target_id, str)
        assert isinstance(etype, str)
        assert isinstance(eprop_str, str)
        assert len(source_id) > 0
        assert len(target_id) > 0
        assert len(etype) > 0
        assert len(eprop_str) > 0
        eprop = json.loads(eprop_str)
        assert isinstance(eprop, dict)

    # Comparisons
    assert num_nodes == len(node_data)
    assert num_edges == len(edge_data)


def test_file_jsonl(workdir, project_params):
    project_name, project_class, project_version, num_nodes, num_edges = project_params
    project = project_class(project_version, workdir)
    project.to_jsonl()
    kgw.run([project])

    # File checks
    project_dir = utils.get_project_dir(project_name, project_version)
    filepath_nodes = os.path.join(workdir, project_dir, "results", "kg_nodes.jsonl")
    filepath_edges = os.path.join(workdir, project_dir, "results", "kg_edges.jsonl")
    utils.check_file(filepath_nodes)
    utils.check_file(filepath_edges)

    # Content checks
    node_data = utils.load_jsonl_file(filepath_nodes)
    for row in node_data:
        nid = row["id"]
        ntype = row["type"]
        nprop = row["properties"]
        assert isinstance(nid, str)
        assert isinstance(ntype, str)
        assert isinstance(nprop, dict)
        assert len(nid) > 0
        assert len(ntype) > 0
        assert len(nprop) >= 0  # may be an empty dict

    edge_data = utils.load_jsonl_file(filepath_edges)
    for row in edge_data:
        source_id = row["source_id"]
        target_id = row["target_id"]
        etype = row["type"]
        eprop = row["properties"]
        assert isinstance(source_id, str)
        assert isinstance(target_id, str)
        assert isinstance(etype, str)
        assert isinstance(eprop, dict)
        assert len(source_id) > 0
        assert len(target_id) > 0
        assert len(etype) > 0
        assert len(eprop) >= 0  # may be an empty dict

    # Comparisons
    assert num_nodes == len(node_data)
    assert num_edges == len(edge_data)


def test_file_graphml(workdir, project_params):
    project_name, project_class, project_version, num_nodes, num_edges = project_params
    project = project_class(project_version, workdir)
    project.to_graphml()
    kgw.run([project])

    # File checks
    project_dir = utils.get_project_dir(project_name, project_version)
    filepath = os.path.join(workdir, project_dir, "results", "kg.graphml")
    utils.check_file(filepath)

    # Content checks
    num_nodes_ig, num_edges_ig, parsed_nodes_ig = utils.load_graphml_igraph(filepath)
    num_nodes_nx, num_edges_nx, parsed_nodes_nx = utils.load_graphml_networkx(filepath)

    # Comparisons
    assert num_nodes == num_nodes_ig == num_nodes_nx
    assert num_edges == num_edges_ig == num_edges_nx
    assert parsed_nodes_ig == parsed_nodes_nx


@pytest.mark.parametrize(
    "representation", ["spo", "properties_aggregated", "properties_expanded"]
)
def test_file_metta(workdir, project_params, representation):
    project_name, project_class, project_version, num_nodes, num_edges = project_params
    project = project_class(project_version, workdir)
    project.to_metta(representation=representation)
    kgw.run([project])

    # File checks
    project_dir = utils.get_project_dir(project_name, project_version)
    filepath = os.path.join(
        workdir, project_dir, "results", f"kg_{representation}.metta"
    )
    utils.check_file(filepath)

    # Content checks
    with open(filepath) as f:
        previous_number = 0
        for cnt_line, line in enumerate(f):
            # Skip comment lines
            if line.startswith(";"):
                continue

            # Ensure line starts and ends with a parenthesis
            assert line.startswith("(")
            assert line.endswith(")\n")

            # Ensure number stays same or is incremented by 1 from line to line
            if representation == "properties_expanded":
                number = int(line.split(" ", 1)[0][1:])
                assert number == previous_number or number == (previous_number + 1)
                previous_number = number

            # Ensure last parenthesis closes the first one, i.e. not depth 0 in between
            depth = 0
            idx_first_parenthesis = 0
            idx_last_parenthesis = (
                len(line) - 2
            )  # -2 instead of -1 due to \n being last character
            pause_parenthesis_tracking = False
            cnt_previous_backslashes = 0
            for i, c in enumerate(line):
                if depth == 0:
                    assert (
                        i <= idx_first_parenthesis or i >= idx_last_parenthesis
                    ), f"Invalid parentheses in {filepath} on line {cnt_line} at character {i}"
                if c == '"' and cnt_previous_backslashes % 2 == 0:
                    # ignore parenthesis inside string literals, which start and end
                    # with an unescaped " character
                    pause_parenthesis_tracking = not pause_parenthesis_tracking
                if not pause_parenthesis_tracking and c == "(":
                    depth += 1
                if not pause_parenthesis_tracking and c == ")":
                    depth -= 1
                if c == "\\":
                    cnt_previous_backslashes += 1
                else:
                    cnt_previous_backslashes = 0
            assert (
                depth == 0
            ), f"Imbalanced parentheses in {filepath} on line {cnt_line} at character {i}"



import logging
import os
from contextlib import contextmanager

import luigi

from . import extract, load


# Contextmanager for temporary files


@contextmanager
def temp_output_path(filepath, suffix=".partial", reuse_partial_file=False):
    temp_filepath = filepath + suffix
    if os.path.exists(temp_filepath):
        if not reuse_partial_file:
            os.remove(temp_filepath)
    try:
        # Yield the temp file path for use in the block
        yield temp_filepath
        # Once the block completes, rename temp file to final file path
        os.rename(temp_filepath, filepath)
    except Exception as excp:
        if not reuse_partial_file:
            if os.path.exists(temp_filepath):
                os.remove(temp_filepath)
        raise excp


# Reporting behavior for all tasks


class ReportingTask(luigi.Task):
    logger = logging.getLogger("kgw")
    RED = "\033[91m"
    GREEN = "\033[92m"
    BLUE = "\033[94m"
    RESET = "\033[0m"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Prevent occasionally duplicated log messages
        self._start_happened = False
        self._success_happened = False
        self._failure_happened = False

    @luigi.Task.event_handler(luigi.Event.START)
    def on_start(self):
        if not self._start_happened:
            self.logger.info(f"{self.BLUE}Started{self.RESET}   {self}")
        self._start_happened = True

    @luigi.Task.event_handler(luigi.Event.SUCCESS)
    def on_success(self):
        if not self._success_happened:
            self.logger.info(f"{self.GREEN}Finished{self.RESET}  {self}")
        self._success_happened = True

    @luigi.Task.event_handler(luigi.Event.FAILURE)
    def on_failure(self, exception):
        excp_name = type(exception).__name__
        excp_text = str(exception)
        if not self._failure_happened:
            self.logger.info(
                f"{self.RED}Failed{self.RESET}    {self}\n"
                f"                               {excp_name}: {excp_text}"
            )
        self._failure_happened = True


# Extract


class DirectoryTarget(luigi.Target):
    def __init__(self, dirpath):
        self.dirpath = dirpath

    def exists(self):
        return os.path.isdir(self.dirpath)

    def touch(self):
        os.makedirs(self.dirpath, exist_ok=True)


class CreateDirectory(ReportingTask):
    dirpath = luigi.Parameter()

    def output(self):
        return DirectoryTarget(self.dirpath)

    def run(self):
        self.output().touch()


class DownloadFile(ReportingTask):
    dirpath = luigi.Parameter()
    filename = luigi.Parameter()
    url = luigi.Parameter(significant=False)
    md5 = luigi.Parameter(default="", significant=False)
    sha256 = luigi.Parameter(default="", significant=False)

    # https://luigi.readthedocs.io/en/stable/configuration.html#per-task-retry-policy
    retry_count = 3

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filepath = os.path.join(self.dirpath, self.filename)

    def requires(self):
        return CreateDirectory(self.dirpath)

    def output(self):
        return luigi.LocalTarget(self.filepath)

    def run(self):
        with temp_output_path(self.filepath, reuse_partial_file=True) as filepath_tmp:
            extract.fetch_file(self.url, filepath_tmp)
            if self.md5 != "":
                if not extract.is_valid_file_by_md5(filepath_tmp, self.md5):
                    message = f"Invalid file (MD5 hash does not match): {self.filepath}"
                    raise Exception(message)
            if self.sha256 != "":
                if not extract.is_valid_file_by_sha256(filepath_tmp, self.sha256):
                    message = (
                        f"Invalid file (SHA256 hash does not match): {self.filepath}"
                    )
                    raise Exception(message)


# Load


class CreateStatisticsFile(ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()
    transform_task = luigi.TaskParameter(significant=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "results")
        self.filepath = os.path.join(self.subdirpath, "statistics.json")

    def requires(self):
        tasks = {
            "subdirpath": CreateDirectory(self.subdirpath),
            "sqlite_file": self.transform_task(self.dirpath, self.version),
        }
        return tasks

    def output(self):
        return luigi.LocalTarget(self.filepath)

    def run(self):
        db_filepath = self.input()["sqlite_file"].path
        with temp_output_path(self.filepath) as statistics_filepath:
            load.sqlite_to_statistics(db_filepath, statistics_filepath)


class CreateSchemaFile(ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()
    node_type_to_color = luigi.DictParameter(significant=False)
    transform_task = luigi.TaskParameter(significant=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "results")
        self.filepath = os.path.join(self.subdirpath, "schema.html")

    def requires(self):
        tasks = {
            "subdirpath": CreateDirectory(self.subdirpath),
            "sqlite_file": self.transform_task(self.dirpath, self.version),
        }
        return tasks

    def output(self):
        return luigi.LocalTarget(self.filepath)

    def run(self):
        db_filepath = self.input()["sqlite_file"].path
        with temp_output_path(self.filepath) as schema_filepath:
            load.sqlite_to_schema(db_filepath, schema_filepath, self.node_type_to_color)


class CreateCompactSchemaFile(ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()
    node_type_to_color = luigi.DictParameter(significant=False)
    transform_task = luigi.TaskParameter(significant=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "results")
        self.filepath = os.path.join(self.subdirpath, "schema.html")

    def requires(self):
        tasks = {
            "subdirpath": CreateDirectory(self.subdirpath),
            "sqlite_file": self.transform_task(self.dirpath, self.version),
        }
        return tasks

    def output(self):
        return luigi.LocalTarget(self.filepath)

    def run(self):
        db_filepath = self.input()["sqlite_file"].path
        with temp_output_path(self.filepath) as schema_filepath:
            load.sqlite_to_schema_compact(
                db_filepath, schema_filepath, self.node_type_to_color
            )


class CreateSqlFile(ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()
    transform_task = luigi.TaskParameter(significant=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "results")
        self.filepath = os.path.join(self.subdirpath, "kg.sql")

    def requires(self):
        tasks = {
            "subdirpath": CreateDirectory(self.subdirpath),
            "sqlite_file": self.transform_task(self.dirpath, self.version),
        }
        return tasks

    def output(self):
        return luigi.LocalTarget(self.filepath)

    def run(self):
        db_filepath = self.input()["sqlite_file"].path
        with temp_output_path(self.filepath) as sql_filepath:
            load.sqlite_to_sql(db_filepath, sql_filepath)


class CreateCsvNodesFile(ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()
    transform_task = luigi.TaskParameter(significant=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "results")
        self.filepath = os.path.join(self.subdirpath, "kg_nodes.csv")

    def requires(self):
        tasks = {
            "subdirpath": CreateDirectory(self.subdirpath),
            "sqlite_file": self.transform_task(self.dirpath, self.version),
        }
        return tasks

    def output(self):
        return luigi.LocalTarget(self.filepath)

    def run(self):
        db_filepath = self.input()["sqlite_file"].path
        with temp_output_path(self.filepath) as csv_filepath:
            load.sqlite_to_csv(db_filepath, csv_filepath, "nodes")


class CreateCsvEdgesFile(ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()
    transform_task = luigi.TaskParameter(significant=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "results")
        self.filepath = os.path.join(self.subdirpath, "kg_edges.csv")

    def requires(self):
        tasks = {
            "subdirpath": CreateDirectory(self.subdirpath),
            "sqlite_file": self.transform_task(self.dirpath, self.version),
        }
        return tasks

    def output(self):
        return luigi.LocalTarget(self.filepath)

    def run(self):
        db_filepath = self.input()["sqlite_file"].path
        with temp_output_path(self.filepath) as csv_filepath:
            load.sqlite_to_csv(db_filepath, csv_filepath, "edges")


class CreateJsonlNodesFile(ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()
    transform_task = luigi.TaskParameter(significant=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "results")
        self.filepath = os.path.join(self.subdirpath, "kg_nodes.jsonl")

    def requires(self):
        tasks = {
            "subdirpath": CreateDirectory(self.subdirpath),
            "sqlite_file": self.transform_task(self.dirpath, self.version),
        }
        return tasks

    def output(self):
        return luigi.LocalTarget(self.filepath)

    def run(self):
        db_filepath = self.input()["sqlite_file"].path
        with temp_output_path(self.filepath) as jsonl_filepath:
            load.sqlite_to_jsonl(db_filepath, jsonl_filepath, "nodes")


class CreateJsonlEdgesFile(ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()
    transform_task = luigi.TaskParameter(significant=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "results")
        self.filepath = os.path.join(self.subdirpath, "kg_edges.jsonl")

    def requires(self):
        tasks = {
            "subdirpath": CreateDirectory(self.subdirpath),
            "sqlite_file": self.transform_task(self.dirpath, self.version),
        }
        return tasks

    def output(self):
        return luigi.LocalTarget(self.filepath)

    def run(self):
        db_filepath = self.input()["sqlite_file"].path
        with temp_output_path(self.filepath) as jsonl_filepath:
            load.sqlite_to_jsonl(db_filepath, jsonl_filepath, "edges")


class CreateMettaFile(ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()
    representation = luigi.Parameter()
    transform_task = luigi.TaskParameter(significant=False)
    batch_size = luigi.IntParameter(default=10_000, significant=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "results")
        self.filepath = os.path.join(self.subdirpath, f"kg_{self.representation}.metta")

    def requires(self):
        tasks = {
            "subdirpath": CreateDirectory(self.subdirpath),
            "sqlite_file": self.transform_task(self.dirpath, self.version),
        }
        return tasks

    def output(self):
        return luigi.LocalTarget(self.filepath)

    def run(self):
        db_filepath = self.input()["sqlite_file"].path
        with temp_output_path(self.filepath) as metta_filepath:
            if self.representation == "spo":
                load.sqlite_to_metta_repr1(db_filepath, metta_filepath, self.batch_size)
            elif self.representation == "properties_aggregated":
                load.sqlite_to_metta_repr2(db_filepath, metta_filepath, self.batch_size)
            elif self.representation == "properties_expanded":
                load.sqlite_to_metta_repr3(db_filepath, metta_filepath, self.batch_size)
            else:
                message = (
                    f"Invalid representation: {self.representation}\n"
                    'Available options: "rpo", "properties_aggregated", "properties_expanded"'
                )
                raise ValueError(message)


class CreateGraphMlFile(ReportingTask):
    dirpath = luigi.Parameter()
    version = luigi.Parameter()
    transform_task = luigi.TaskParameter(significant=False)
    batch_size = luigi.IntParameter(default=10_000, significant=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.subdirpath = os.path.join(self.dirpath, "results")
        self.filepath = os.path.join(self.subdirpath, "kg.graphml")

    def requires(self):
        tasks = {
            "subdirpath": CreateDirectory(self.subdirpath),
            "sqlite_file": self.transform_task(self.dirpath, self.version),
        }
        return tasks

    def output(self):
        return luigi.LocalTarget(self.filepath)

    def run(self):
        db_filepath = self.input()["sqlite_file"].path
        with temp_output_path(self.filepath) as graphml_filepath:
            load.sqlite_to_graphml(db_filepath, graphml_filepath, self.batch_size)
